#!/usr/bin/env python


import fsleyes_props as props

class Base(props.Props):
    thing = props.Choice((1, 2, 3))


class Sub(Base):
    thing = props.Choice((4, 5, 6))

# fsl/fsleyes/props!58
def test_property_overwritten():
    b = Base()
    s = Sub()

    assert Base.thing.getChoices()  == [1, 2, 3]
    assert Base.thing.getChoices(b) == [1, 2, 3]
    assert Sub .thing.getChoices()  == [4, 5, 6]
    assert Sub .thing.getChoices(s) == [4, 5, 6]



#!/usr/bin/env python
#
# suppress.py - Notification suppression.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides a couple of functions which may be used as context
managers for suppressing notification of changes to :class:`.HasProperties`
property values.

.. autosummary::
   :nosignatures:

   suppress
   suppressAll
   skip
"""


import contextlib


@contextlib.contextmanager
def suppress(hasProps, *propNames, notify=False):
    """Suppress notification for the given properties on the given
    :class:`.HasProperties` instance.

    This function saves the current notification state of the properties,
    disables notification, yields to the calling code, and then restores the
    notification state.

    :arg hasProps: ``HasProperties`` instance.

    :arg propNames: Properties to suppress notifications for.

    :arg notify:    If ``True``, a notification will be triggered
                    on ``propNames`` via :meth:`.HasProperties.propNotify`,
                    exit. Defaults to ``False``.

    This function is intended to be used as follows::

        with suppress(hasProps1, 'prop1'):
            # Do stuff which might cause unwanted
            # property value notifications to occur
    """

    states = [hasProps.getNotificationState(p) for p in propNames]

    for propName in propNames:
        hasProps.disableNotification(propName)

    try:
        yield

    finally:
        for propName, state in zip(propNames, states):
            hasProps.setNotificationState(propName, state)

    if notify:
        for propName in propNames:
            hasProps.propNotify(propName)


@contextlib.contextmanager
def suppressAll(hasProps):
    """Suppress notification for all properties on the given
    :class:`.HasProperties` instance.

    :arg hasProps: The ``HasProperties`` instance to suppress.


    This function is intended to be used as follows::

        with suppressAll(hasProps):
            # Do stuff which might cause unwanted
            # property value notifications to occur


    .. note:: After this function has completed. notification
              will be enabled for all properties of ``hasProps``.
              This is in contrast to the :func:`suppress` function,
              which restores the previous notification state of
              the property.
    """

    hasProps.disableAllNotification()

    try:
        yield
    finally:
        hasProps.enableAllNotification()


@contextlib.contextmanager
def skip(hasProps, propNames, listenerName, ignoreInvalid=False):
    """Skip the listener with the specified listener name, if/when
    changes to the specified property/properties occur.

    :arg hasProps:      A :class:`.HasProperties` instance.
    :arg propNames:     Name of a property on ``hasProps``, or a sequence of
                        property names.
    :arg listenerName:  Name of the listener to skip.
    :arg ignoreInvalid: Defaults to ``False``. If ``True``, passing a
                        non-existent listener will not result in an error.
    """

    if isinstance(propNames, str):
        propNames = [propNames]

    listenerStates = {}

    for propName in propNames:
        if (not ignoreInvalid) or \
           hasProps.hasListener(propName, listenerName):
            listenerStates[propName] = hasProps.getListenerState(
                propName, listenerName)
            hasProps.disableListener(propName, listenerName)

    try:
        yield

    finally:
        for propName, state in listenerStates.items():
            hasProps.setListenerState(propName, listenerName, state)



#!/usr/bin/env python
#
# widgets_boolean.py - Generate a widget to control a Boolean property.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :func:`_Boolean` function, which is imported into
the :mod:`widgets` module namespace. It is separated purely to keep the
``widgets`` module file size down.
"""


import wx

from . import                          widgets
import fsleyes_widgets.bitmapradio  as bmpradio
import fsleyes_widgets.bitmaptoggle as bmptoggle


def _Boolean(parent,
             hasProps,
             propObj,
             propVal,
             icon=None,
             toggle=True,
             style=None,
             **kwargs):
    """Creates and returns a ``wx.CheckBox``, allowing the user to set the
    given :class:`.Boolean` property value.

    :arg icon:   If the ``icon`` argument is provided, it should be a string
                 containing the name of an image file, or a list of two image
                 file names.  In this case, case, a
                 `:class:`.BitmapToggleButton` is used instead of a
                 ``CheckBox``.

                 .. note:: If ``toggle`` is ``False``, ``icons`` may
                           alternately be a sequence of four icons, with the
                           order ``[selectedTrueIcon, unselectedTrueIcon,
                           selectedFalseIcon, unselectedFalseIcon]``.

    :arg toggle: If two icon images are provided, and the ``toggle`` argument
                 is ``True`` (the default), a :class:`.BitmapToggleButton` is
                 used. If ``toggle=False``, a :class:`.BitmapRadioBox` is used
                 instead.  In the latter case, the ``style`` argument is
                 passed through to the :meth:`.BitmapRadioBox.__init__`
                 method.

    :arg style:  If ``toggle`` is ``False``, this value is passed through to
                 the :meth:`BitmapRadioBox.__init__` method.

    See the :func:`.widgets._String` documentation for details on the other
    parameters.
    """

    if icon is None:
        widget, event, wget, wset = _booleanCheckBox(parent, **kwargs)

    else:

        if isinstance(icon, str):
            icon = [icon]

        for i in range(len(icon)):
            icon[i] = wx.Bitmap(icon[i], wx.BITMAP_TYPE_PNG)

        if toggle:
            widget, event, wget, wset = _booleanToggle(parent, icon)
        else:
            widget, event, wget, wset = _booleanRadio( parent, icon, style)

    widgets._propBind(hasProps,
                      propObj,
                      propVal,
                      widget,
                      event,
                      widgetGet=wget,
                      widgetSet=wset)
    return widget


def _booleanCheckBox(parent, cblabel=None):
    """Create a ``wx.CheckBox`` to link to the :class:`.Boolean` property.

    :arg cblabel: Label to show alongside the checkbox. Named so because
                  ``label`` is absorbed by the :class:`.ViewItem` class.
    """

    kwargs = {}
    if cblabel is not None:
        kwargs['label'] = cblabel
    widget = wx.CheckBox(parent, **kwargs)
    event  = wx.EVT_CHECKBOX
    widget.SetMinSize(widget.GetBestSize())
    return widget, event, None, None


def _booleanToggle(parent, icons):
    """Create a :class:`.BitmapToggleButton` to link to the :class:`.Boolean`
    property.
    """
    if len(icons) == 1:
        icons = icons + [None]

    trueBmp  = icons[0]
    falseBmp = icons[1]

    style  = wx.BU_EXACTFIT | wx.ALIGN_CENTRE | wx.BU_NOTEXT
    widget = bmptoggle.BitmapToggleButton(
        parent,
        trueBmp=trueBmp,
        falseBmp=falseBmp,
        style=style)
    event  = bmptoggle.EVT_BITMAP_TOGGLE
    widget.SetMinSize(widget.GetBestSize())

    return widget, event, None, None


def _booleanRadio(parent, icons, style):
    """Create a :class:`.BitmapRadioBox` to link to the :class:`.Boolean`
    property.
    """

    widget = bmpradio.BitmapRadioBox(parent, style)
    event  = bmpradio.EVT_BITMAP_RADIO_EVENT
    widget.SetMinSize(widget.GetBestSize())

    if len(icons) == 2:
        icons = [icons[0], None, icons[1], None]

    widget.AddChoice(icons[0], unselectedBmp=icons[1])
    widget.AddChoice(icons[2], unselectedBmp=icons[3])

    def wget():
        return widget.GetSelection() == 0

    def wset(val):
        if val: widget.SetSelection(0)
        else:   widget.SetSelection(1)

    return widget, event, wget, wset



#!/usr/bin/env python
#
# test_propertyvalue.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import fsleyes_props.properties_value as properties_value

import logging

logging.basicConfig()
logging.getLogger('fsleyes_props').setLevel(logging.DEBUG)

class Context:
    pass


def test_listener():

    ctx    = Context()
    pv     = properties_value.PropertyValue(ctx)
    called = {}

    # TODO Test args passed to listener

    def l1(*a):
        called['l1'] = called.get('l1', 0) + 1

    def l2():
        called['l2'] = called.get('l2', 0) + 1

    def l3(a, b, c, d):
        called['l3'] = called.get('l3', 0) + 1

    def l7(a, b, *rest):
        called['l7'] = called.get('l7', 0) + 1

    def al1(*a):
        called['al1'] = called.get('al1', 0) + 1

    def al2():
        called['al2'] = called.get('al2', 0) + 1

    def al3(a, b, c, d):
        called['al3'] = called.get('al3', 0) + 1

    def al7(a, b, *rest):
        called['al7'] = called.get('al7', 0) + 1

    class ListenerObj:

        def l4(self, *a):
            called['l4'] = called.get('l4', 0) + 1

        def l5(self):
            called['l5'] = called.get('l5', 0) + 1

        def l6(self, a, b, c, d):
            called['l6'] = called.get('l6', 0) + 1

        def al4(self, *a):
            called['al4'] = called.get('al4', 0) + 1

        def al5(self):
            called['al5'] = called.get('al5', 0) + 1

        def al6(self, a, b, c, d):
            called['al6'] = called.get('al6', 0) + 1

    lobj = ListenerObj()

    pv.addListener(         'l1',  l1)
    pv.addListener(         'l2',  l2)
    pv.addListener(         'l3',  l3)
    pv.addListener(         'l4',  lobj.l4)
    pv.addListener(         'l5',  lobj.l5)
    pv.addListener(         'l6',  lobj.l6)
    pv.addListener(         'l7',  l7)

    pv.addAttributeListener('al1', al1)
    pv.addAttributeListener('al2', al2)
    pv.addAttributeListener('al3', al3)
    pv.addAttributeListener('al4', lobj.al4)
    pv.addAttributeListener('al5', lobj.al5)
    pv.addAttributeListener('al6', lobj.al6)
    pv.addAttributeListener('al7', al7)

    listeners    = ['l1', 'l2', 'l3', 'l4', 'l5', 'l6', 'l7']
    attListeners = ['al1', 'al2', 'al3', 'al4', 'al5', 'al6', 'al7']

    for l in listeners:
        assert pv.hasListener(l)

    pv.set('New value')
    for l in listeners:
        assert called[l] == 1

    # value unchanged, notification should not take place
    pv.set('New value')
    for l in listeners:
        assert called[l] == 1

    pv.set('New new value')
    for l in listeners:
        assert called[l] == 2

    for l in attListeners:
        assert called.get(l) is None

    pv.setAttribute('newAtt', 'value')
    for l in attListeners:
        assert called[l] == 1

    pv.setAttribute('newAtt', 'value')
    for l in attListeners:
        assert called[l] == 1

    pv.setAttribute('newAtt', 'new value')
    for l in attListeners:
        assert called[l] == 2



#!/usr/bin/env python
#
# test_hasprops.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import fsleyes_props as props


# fsl/fsleyes/props!63
def test_self_reference():

    class Thing(props.HasProps):
        ref = props.Object()

    t = Thing()

    vals = [None, 123, 'abc', t]

    for v in vals:
        t.ref = v
        print(t)



#!/usr/bin/env python
#
# cli.py - Generate command line arguments for a HasProperties object.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

"""Generate command line arguments for a :class:`.HasProperties` instance.

This module provides the following functions:

 .. autosummary::
    :nosignatures:

    addParserArguments
    applyArguments
    generateArguments

The ``addParserArguments`` function is used to add arguments to an
``ArgumentParser`` object for the properties of a ``HasProperties`` class. The
simplest way to do so is to allow the ``addParserArguments`` function to
automatically generate short and long arguments from the property names::

    >>> import argparse
    >>> import fsleyes_props as props

    >>> class MyObj(props.HasProperties):
            intProp  = props.Int()
            boolProp = props.Boolean()

    >>> parser = argparse.ArgumentParser('MyObj')
    >>> props.addParserArguments(MyObj, parser)

    >>> parser.print_help()
    usage: MyObj [-h] [-b] [-i INT]

    optional arguments:
        -h, --help            show this help message and exit
        -b, --boolProp
        -i INT, --intProp INT

Now, if we have a ``MyObj`` instance, and some arguments::

    >>> myobj = MyObj()

    >>> args = parser.parse_args(['-b', '--intProp', '52'])

    >>> print myobj
    MyObj
      boolProp = False
       intProp = 0

    >>> props.applyArguments(myobj, args)
    >>> print myobj
    MyObj
      boolProp = True
       intProp = 52

If you want to customise the short and long argument tags, and the help text,
for each property, you can pass them in to the ``addParserArguments``
function::

    >>> shortArgs = {'intProp' : 'r',              'boolProp' : 't'}
    >>> longArgs  = {'intProp' : 'TheInt',         'boolProp' : 'someBool'}
    >>> propHelp  = {'intProp' : 'Sets int value', 'boolProp' : 'Toggles bool'}

    >>> parser = argparse.ArgumentParser('MyObj')
    >>> props.addParserArguments(MyObj,
                                 parser,
                                 shortArgs=shortArgs,
                                 longArgs=longArgs,
                                 propHelp=propHelp)
    >>> parser.print_help()
    usage: MyObj [-h] [-t] [-r INT]

    optional arguments:
      -h, --help            show this help message and exit
      -t, --someBool        Toggles bool
      -r INT, --TheInt INT  Sets int value

Or, you can add the short and long arguments, and the help text, as specially
named class attributes of your ``HasProperties`` class or instance::

    >>> class MyObj(props.HasProperties):
            intProp  = props.Int()
            boolProp = props.Boolean()
            _shortArgs = {
                'intProp'  : 'r',
                'boolProp' : 't'
            }
            _longArgs = {
                'intProp'  : 'TheInt',
                'boolProp' : 'someBool'
            }
            _propHelp = {
                'intProp' : 'Sets int value',
                'boolProp' : 'Toggles bool'
            }

    >>> parser = argparse.ArgumentParser('MyObj')
    >>> props.addParserArguments(MyObj, parser)

    >>> parser.print_help()
    usage: MyObj [-h] [-t] [-r INT]

    optional arguments:
      -h, --help            show this help message and exit
      -t, --someBool        Toggles bool
      -r INT, --TheInt INT  Sets int value

    >>> args = parser.parse_args('--someBool -r 23413'.split())
    >>> myobj = MyObj()
    >>> props.applyArguments(myobj, args)
    >>> print myobj
    MyObj
      boolProp = True
       intProp = 23413

The ``generateArguments`` function, as the name suggests, generates command
line arguments from a ``HasProperties`` instance::

    >>> props.generateArguments(myobj)
    ['--someBool', '--TheInt', '23413']

The ``generateArguments`` and ``applyArguments`` functions optionally accept a
set of *transform* functions which, for ``generateArguments``, take the value
of a property, and return some transformation of that property, suitable to be
used as a command line argument value. The transform functions passed to the
``applyArguments`` function perform the reverse transformation.  Transforms
are useful for properties which cannot easily be converted to/from strings,
and also for properties where the value you wish users to pass in on the
command line does not correspond exactly to the value you wish the property to
be given.

For example::

    >>> class MyObject(props.HasProperties):
            showBlah = props.Boolean(default=True)

    >>> shortArgs = {'showBlah' : 'hb'}
    >>> longArgs  = {'showBlah' : 'hideBlah'}
    >>> xforms    = {'showBlah' : lambda b : not b }

    >>> parser = argparse.ArgumentParser('MyObject')
    >>> props.addParserArguments(MyObject,
                                 parser,
                                 shortArgs=shortArgs,
                                 longArgs=longArgs)

    >>> myobj = MyObject()
    >>> myobj.showBlah = False

    >>> props.generateArguments(myobj,
                                shortArgs=shortArgs,
                                longArgs=longArgs,
                                xformFuncs=xforms)
        ['--hideBlah']

In this example, we can use the same transform function for the reverse
operation::

    >>> myobj2 = MyObject()
    >>> args   = parser.parse_args(['--hideBlah'])
    >>> props.applyArguments(myobj2,
                             args,
                             xformFuncs=xforms,
                             longArgs=longArgs)
    >>> print myobj2
        MyObject
            showBlah = False

The ``cli`` module supports the following property types:

.. autosummary::
   :nosignatures:

   _String
   _Choice
   _Boolean
   _Int
   _Real
   _Percentage
   _Bounds
   _Point
   _Colour
   _ColourMap

"""


import logging

import sys
import argparse

import matplotlib        as mpl
import matplotlib.pyplot as plt

import fsleyes_props.properties       as props
import fsleyes_props.properties_types as ptypes


log = logging.getLogger(__name__)


class SkipArgument(Exception):
    """This exception may be raised by transform functions which are called
    by :func:`applyArguments` and :func:`generateArguments` to indicate
    that the arguemnt should be skipped (i.e. not applied, or not generated).
    """


def _String(parser,
            propObj,
            propCls,
            propName,
            propHelp,
            shortArg,
            longArg,
            atype=str):
    """Adds an argument to the given parser for the given :class:`.String`
    property.

    :param parser:   An ``ArgumentParser`` instance.

    :param propCls:  A ``HasProperties`` class or instance.

    :param propObj:  The ``PropertyBase`` class.

    :param propName: Name of the property.

    :param propHelp: Custom help text for the property.

    :param shortArg: String to use as the short argument.

    :param longArg:  String to use as the long argument.

    :param atype:    Data type to expect (defaults to ``str``). This allows
                     the conversion type to be overridden for some, but not
                     all, property types.
    """
    parser.add_argument(shortArg, longArg, help=propHelp, type=atype)


def _Choice(parser,
            propObj,
            propCls,
            propName,
            propHelp,
            shortArg,
            longArg,
            **kwargs):
    """Adds an argument to the given parser for the given :class:`.Choice`
    property. See the :func:`_String` documentation for details on the
    parameters.

    Only works with ``Choice`` properties with string options (unless
    the ``choices`` argument is provided). See the ``allowStr`` parameter
    for :class:`.Choice` instances.

    All of the described arguments must be speified as keyword arguments.
    Any additional arguments will be passed to the
    ``ArgumentParser.add_argument`` method.

    :arg choices: If not ``None``, assumed to be list of possible
                  choices for the property. If not provided, the possible
                  choices are taken from the :meth:`.Choice.getChoices`
                  method. If explicitly passed in as ``None``, the argument
                  will be configured to accept any value.

    :arg default: If not ``None``, gives the default value. Otherwise,
                  the ``default`` attribute of the :class:`.Choice`
                  object is used.

    :arg useAlts: If ``True`` (the default), alternate values for the
                  choices are added as options (see the :class:`.Choice`
                  class). n.b. This flag will have no effect if  ``choices``
                  is explicitly set to ``None``, as desrcibed above.

    :arg metavar: If ``None`` (the default), the available choices for
                  this property will be shown in the help text. Otherwise,
                  this string will be shown instead.
    """

    choices = kwargs.pop('choices', 'notspecified')
    default = kwargs.pop('default', None)
    useAlts = kwargs.pop('useAlts', True)
    metavar = kwargs.pop('metavar', None)

    if choices == 'notspecified':
        choices = propObj.getChoices()

    if useAlts and choices is not None:
        alternates = propObj.getAlternates()

        for altList in alternates:
            choices += [a for a in altList]

    if default is None:
        default = propObj.getAttribute(None, 'default')

    # Make sure that only unique
    # choices are set as options.
    if choices is not None:

        # We could just convert to a set, but
        # this would not preserve ordering, so
        # we'll uniquify the list the hard way.
        choices = [str(c) for c in choices]
        unique  = []

        for c in choices:
            if c not in unique:
                unique.append(c)

        choices = unique

    parser.add_argument(shortArg,
                        longArg,
                        help=propHelp,
                        choices=choices,
                        metavar=metavar,
                        **kwargs)


def _Boolean(parser, propObj, propCls, propName, propHelp, shortArg, longArg):
    """Adds an argument to the given parser for the given :class:`.Boolean`
    property. See the :func:`_String` documentation for details on the
    parameters.
    """
    # Using store_const instead of store_true,
    # because if the user doesn't set this
    # argument, we don't want to explicitly
    # set the property value to False (if it
    # has a default value of True, we don't
    # want that default value overridden).
    parser.add_argument(shortArg,
                        longArg,
                        help=propHelp,
                        action='store_const',
                        const=True)


def _Int(parser, propObj, propCls, propName, propHelp, shortArg, longArg):
    """Adds an argument to the given parser for the given :class:`.Int`
    property. See the :func:`_String` documentation for details on the
    parameters.
    """
    parser.add_argument(shortArg,
                        longArg,
                        help=propHelp,
                        metavar='INT',
                        type=int)


def _Real(parser, propObj, propCls, propName, propHelp, shortArg, longArg):
    """Adds an argument to the given parser for the given :class:`.Real`
    property. See the :func:`_String` documentation for details on the
    parameters.
    """
    parser.add_argument(shortArg,
                        longArg,
                        help=propHelp,
                        metavar='REAL',
                        type=float)


def _Percentage(
        parser, propObj, propCls, propName, propHelp, shortArg, longArg):
    """Adds an argument to the given parser for the given :class:`.Percentage`
    property. See the :func:`_String` documentation for details on the
    parameters.
    """
    parser.add_argument(shortArg,
                        longArg,
                        help=propHelp,
                        metavar='PERC',
                        type=float)


def _Bounds(parser,
            propObj,
            propCls,
            propName,
            propHelp,
            shortArg,
            longArg,
            atype=None):
    """Adds an argument to the given parser for the given :class:`.Bounds`
    property. See the :func:`_String` documentation for details on the
    parameters.
    """
    ndims = getattr(propCls, propName)._ndims
    real  = getattr(propCls, propName)._real

    metavar = tuple(['LO', 'HI'] * ndims)

    if atype is None:
        if real: atype = float
        else:    atype = int

    parser.add_argument(shortArg,
                        longArg,
                        help=propHelp,
                        metavar=metavar,
                        type=atype,
                        nargs=2 * ndims)


def _Point(parser, propObj, propCls, propName, propHelp, shortArg, longArg):
    """Adds an argument to the given parser for the given :class:`.Point`
    property. See the :func:`_String` documentation for details on the
    parameters.
    """
    ndims = getattr(propCls, propName)._ndims
    real  = getattr(propCls, propName)._real
    if real: pType = float
    else:    pType = int
    parser.add_argument(shortArg,
                        longArg,
                        help=propHelp,
                        metavar='N',
                        type=pType,
                        nargs=ndims)


def _Colour(parser,
            propObj,
            propCls,
            propName,
            propHelp,
            shortArg,
            longArg,
            alpha=True):
    """Adds an argument to the given parser for the given :class:`.Colour`
    property. See the :func:`_String` documentation for details on the
    parameters.
    """
    if alpha:
        metavar = ('R', 'G', 'B', 'A')
        nargs   = 4
    else:
        metavar = ('R', 'G', 'B')
        nargs   = 3
    parser.add_argument(shortArg,
                        longArg,
                        help=propHelp,
                        metavar=metavar,
                        type=float,
                        nargs=nargs)


def _ColourMap(parser,
               propObj,
               propCls,
               propName,
               propHelp,
               shortArg,
               longArg,
               parseStr=False,
               choices=None,
               metavar=None):
    """Adds an argument to the given parser for the given :class:`.ColourMap`
    property.

    :arg parseStr: If ``False`` (the default), the parser will be configured
                   to parse any registered ``matplotlib`` colour map name.
                   Otherwise, the parser will accept any string, and assume
                   that the ``ColourMap`` property is set up to accept it.

    :arg choices:  This parameter can be used to restrict the values that
                   will be accepted.

    :arg metavar:  If ``choices`` is not ``None``, they will be shown in the
                   help text, unless this string is provided.

    See the :func:`_String` documentation for details on the other parameters.
    """

    # Attempt to retrieve a matplotlib.cm.ColourMap
    # instance given its name
    def parse(cmapName):
        try:

            cmapKeys   = plt.colormaps()
            cmapNames  = [mpl.colormaps[cm].name for cm in cmapKeys]

            lCmapNames = [s.lower() for s in cmapNames]
            lCmapKeys  = [s.lower() for s in cmapKeys]
            cmapName   = cmapName.lower()

            try:               idx = lCmapKeys .index(cmapName)
            except ValueError: idx = lCmapNames.index(cmapName)

            cmapName = cmapKeys[idx]

            return mpl.colormaps[cmapName]

        except Exception:
            raise argparse.ArgumentTypeError(
                'Unknown colour map: {}'.format(cmapName))

    if choices is not None:
        choices = [c.lower() for c in choices]

    if parseStr: aType = str.lower
    else:        aType = parse

    parser.add_argument(shortArg,
                        longArg,
                        help=propHelp,
                        type=aType,
                        choices=choices,
                        metavar=metavar)


def _getShortArgs(propCls, propNames, exclude=''):
    """Generates unique single-letter argument names for each of the names in
    the given list of property names. Any letters in the exclude string are
    not used as short arguments.

    :param propCls:   A ``HasProperties`` class.

    :param propNames: List of property names for which short arguments
                      are to be generated.

    :param exclude:   String containing letters which should not be used.
    """

    letters   = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
    shortArgs = {}

    for propName in propNames:

        # if '_shortArgs' is present on the hasProps
        # object, and there is an entry for the
        # current property, use that entry.
        if hasattr(propCls, '_shortArgs'):
            if propName in propCls._shortArgs:

                # throw an error if that entry
                # has already been used, or
                # should be excluded
                if propCls._shortArgs[propName] in list(shortArgs.values()) or\
                   propCls._shortArgs[propName] in exclude:
                    raise RuntimeError(
                        'Duplicate or excluded short argument for property '
                        '{}.{}: {}'.format(
                            propCls.__name__,
                            propName,
                            propCls._shortArgs[propName]))

                shortArgs[propName] = propCls._shortArgs[propName]
                continue

        # use the first unique letter in the
        # property name or, if that doesn't
        # work, in the alphabet
        for c in propName + letters:

            if (c not in list(shortArgs.values())) and (c not in exclude):
                shortArgs[propName] = c
                break

    if any([name not in list(shortArgs) for name in propNames]):
        raise RuntimeError('Could not generate default short arguments '
                           'for HasProperties object {} - please provide '
                           'custom short arguments via a _shortArgs '
                           'attribute'.format(propCls.__name__))

    return shortArgs


def applyArguments(hasProps,
                   arguments,
                   propNames=None,
                   xformFuncs=None,
                   longArgs=None,
                   **kwargs):
    """Apply arguments to a ``HasProperties`` instance.

    Given a ``HasProperties`` instance and an ``argparse.Namespace`` instance,
    sets the property values of the ``HasProperties`` instance from the values
    stored in the ``Namespace`` object.

    :param hasProps:   The ``HasProperties`` instance.

    :param arguments:  The ``Namespace`` instance.

    :param propNames:  List of property names to apply. If ``None``, an attempt
                       is made to set all properties. If not ``None``, the
                       property values are set in the order specified by this
                       list.

    :param xformFuncs: A dictionary of ``{property name -> function}``
                       mappings, which can be used to transform the value given
                       on the command line before it is assigned to the
                       property.

    :param longArgs:   Dict containing ``{property name : longArg}`` mappings.

    All other keyword arguments are passed through to the ``xformFuncs``
    functions.
    """

    if propNames is None:
        propNames, propObjs = hasProps.getAllProperties()
    else:
        propObjs = [hasProps.getProp(name) for name in propNames]

    if longArgs is None:
        if hasattr(hasProps, '_longArgs'): longArgs = hasProps._longArgs
        else:                              longArgs = dict(zip(propNames,
                                                               propNames))

    if xformFuncs is None:
        xformFuncs = {}

    for propName, propObj in zip(propNames, propObjs):

        xform   = xformFuncs.get(propName, lambda v, **kwa : v)
        argName = longArgs.get(propName, propName)
        argVal  = getattr(arguments, argName, None)

        # An argVal of None means that no value
        # was passed in for this property
        if argVal is None:
            continue

        try:
            argVal = xform(argVal, **kwargs)
        except SkipArgument:
            continue

        log.debug('Setting {}.{} = {}'.format(
            type(hasProps).__name__,
            propName,
            argVal))

        setattr(hasProps, propName, argVal)


def addParserArguments(
        propCls,
        parser,
        cliProps=None,
        shortArgs=None,
        longArgs=None,
        propHelp=None,
        extra=None,
        exclude=''):
    """Adds arguments to the given ``argparse.ArgumentParser`` for the
    properties of the given ``HasProperties`` class or instance.

    :param propCls:        A ``HasProperties`` class. An instance may
                           alternately be passed in.

    :param parser:         An ``ArgumentParser`` to add arguments to.

    :param list cliProps:  List containing the names of properties to add
                           arguments for. If ``None``, and an attribute called
                           ``_cliProps``' is present on the ``propCls`` class,
                           the value of that attribute is used. Otherwise an
                           argument is added for all properties.

    :param dict shortArgs: Dict containing ``{propName: shortArg}`` mappings,
                           to be used as the short (typically single letter)
                           argument flag for each property. If ``None``, and
                           an attribute called ``_shortArgs`` is present on
                           the ``propCls`` class, the value of that attribute
                           is used. Otherwise, short arguments are
                           automatically generated for each property.

    :param dict longArgs:  Dict containing ``{propName: longArg}`` mappings,
                           to be used as the long argument flag for each
                           property. If ``None``, and an attribute called
                           ``_longArgs`` is present on the ``propCls`` class,
                           the value of that attribute is used. Otherwise, the
                           name of each property is used as its long argument.

    :param dict propHelp:  Dict containing ``{propName: helpString]``
                           mappings, to be used as the help text for each
                           property. If ``None``, and an attribute called
                           ``_propHelp`` is present on the ``propCls`` class,
                           the value of that attribute is used. Otherwise, no
                           help string is used.

    :param dict extra:     Any property-specific settings to be passed through
                           to the parser configuration function (see e.g. the
                           :func:`_Choice` function). If ``None``, and an
                           attribute called ``_propExtra`` is present on the
                           ``propCls`` class, the value of that attribute is
                           used instead.

    :param str exclude:    String containing letters which should not be used
                           as short arguments.
    """

    if isinstance(propCls, props.HasProperties):
        propCls = propCls.__class__

    if cliProps is None:
        if hasattr(propCls, '_cliProps'):
            cliProps = propCls._cliProps
        else:
            cliProps = propCls.getAllProperties()[0]

    if propHelp is None:
        if hasattr(propCls, '_propHelp'): propHelp = propCls._propHelp
        else:                             propHelp = {}

    if longArgs is None:
        if hasattr(propCls, '_longArgs'): longArgs = propCls._longArgs
        else:                             longArgs = dict(zip(cliProps,
                                                              cliProps))

    if shortArgs is None:
        if hasattr(propCls, '_shortArgs'):
            shortArgs = propCls._shortArgs
        else:
            shortArgs = _getShortArgs(propCls, cliProps, exclude)

    if extra is None:
        if hasattr(propCls, '_propExtra'):
            extra = propCls._propExtra
        else:
            extra = {prop : {} for prop in cliProps}

    for propName in cliProps:

        propObj    = propCls.getProp(propName)
        propType   = propObj.__class__.__name__
        parserFunc = getattr(
            sys.modules[__name__],
            '_{}'.format(propType), None)

        if parserFunc is None:
            log.warning('No CLI parser function for property {} '
                        '(type {})'.format(propName, propType))
            continue

        shortArg  =  '-{}'.format(shortArgs[propName])
        longArg   = '--{}'.format(longArgs[ propName])
        propExtra = extra.get(propName, {})

        parserFunc(parser,
                   propObj,
                   propCls,
                   propName,
                   propHelp.get(propName, None),
                   shortArg,
                   longArg,
                   **propExtra)


def generateArguments(hasProps,
                      useShortArgs=False,
                      xformFuncs=None,
                      cliProps=None,
                      shortArgs=None,
                      longArgs=None,
                      exclude='',
                      **kwargs):
    """Given a ``HasProperties`` instance, generates a list of arguments which
    could be used to configure another instance in the same way.

    :param hasProps:     The ``HasProperties`` instance.

    :param useShortArgs: If ``True`` the short argument version is used instead
                         of the long argument version.

    :param xformFuncs:   A dictionary of ``{property name -> function}``
                         mappings, which can be used to perform some arbitrary
                         transformation of property values.


    All other keyword arguments are passed through to the ``xformFuncs``
    functions.

    See the :func:`addParserArguments` function for descriptions of the other
    parameters.
    """
    args = []

    if cliProps is None:
        if hasattr(hasProps, '_cliProps'):
            cliProps = hasProps._cliProps
        else:
            cliProps = hasProps.getAllProperties()[0]

    if longArgs is None:
        if hasattr(hasProps, '_longArgs'): longArgs = hasProps._longArgs
        else:                              longArgs = dict(zip(cliProps,
                                                               cliProps))

    if shortArgs is None:
        if hasattr(hasProps, '_shortArgs'):
            shortArgs = hasProps._shortArgs
        else:
            shortArgs = _getShortArgs(hasProps, cliProps, exclude)

    if xformFuncs is None:
        xformFuncs = {}

    for propName in cliProps:
        propObj = hasProps.getProp(propName)
        xform   = xformFuncs.get(propName, lambda v, **kwa: v)

        try:
            propVal = xform(getattr(hasProps, propName), **kwargs)
        except SkipArgument:
            continue

        # TODO Should I skip a property
        #      if its value is None?
        if propVal is None:
            continue

        if useShortArgs: argKey =  '-{}'.format(shortArgs[propName])
        else:            argKey = '--{}'.format(longArgs[ propName])

        # TODO I should be using the (new)
        #      serialise module for this.
        #      This would incorporate the
        #      TODO below.

        # TODO This logic could somehow be stored
        #      as default transform functions for
        #      the respective types
        if isinstance(propObj, (ptypes.Bounds, ptypes.Point, ptypes.Colour)):
            values = ['{}'.format(v) for v in propVal]

        elif isinstance(propObj, ptypes.ColourMap):
            values = [propVal.name]

        elif isinstance(propObj, ptypes.String):
            values = ['"{}"'.format(propVal)]

        elif isinstance(propObj, ptypes.Boolean):
            values = None
            if not propVal: argKey = None
        else:
            if propVal is None: values = None
            else:               values = [propVal]

        if argKey is not None: args.append(argKey)
        if values is not None: args.extend(values)

    return [str(a) for a in args]



#!/usr/bin/env python
#
# properties_types.py - Definitions for different property types.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""Definitions for different property types.

This module provides a number of :class:`.PropertyBase` subclasses which
define properties of different types. These classes are intended to be
added as attributes of a :class:`.HasProperties` class definition.

 .. autosummary::
    :nosignatures:

    Object
    Boolean
    Number
    Int
    Real
    Percentage
    String
    Choice
    FilePath
    List
    Colour
    ColourMap
    Bounds
    Point
    Array
"""


import os.path as op

import matplotlib        as mpl
import matplotlib.colors as mplcolors
import numpy             as np

from . import properties        as props
from . import properties_value  as propvals


class Object(props.PropertyBase):
    """A property which encapsulates any value. """

    def __init__(self, **kwargs):
        """Create a ``Object`` property. If an ``equalityFunc`` is not
        provided, any writes to this property will be treated as if the value
        has changed (and any listeners will be notified).
        """

        def defaultEquals(this, other):
            return False

        kwargs['equalityFunc'] = kwargs.get('equalityFunc', defaultEquals)
        props.PropertyBase.__init__(self, **kwargs)


class Boolean(props.PropertyBase):
    """A property which encapsulates a ``bool`` value."""

    def __init__(self, **kwargs):
        """Create a ``Boolean`` property.

        If the ``default`` ``kwarg`` is not provided, a default value of
        ``False`` is used.
        """

        kwargs['default'] = kwargs.get('default', False)
        props.PropertyBase.__init__(self, **kwargs)


    def cast(self, instance, attributes, value):
        """Overrides :meth:`.PropertyBase.cast`. Casts the given value to a
        ``bool``.
        """
        return bool(value)


class Number(props.PropertyBase):
    """Base class for the :class:`Int` and :class:`Real` classes.

    A property which represents a number.  Don't use/subclass this,
    use/subclass one of ``Int`` or ``Real``.
    """

    def __init__(self,
                 minval=None,
                 maxval=None,
                 clamped=False,
                 **kwargs):
        """Define a :class:`Number` property.

        :param minval:  Minimum valid value

        :param maxval:  Maximum valid value

        :param clamped: If ``True``, the value will be clamped to its
                        min/max bounds.

        :param kwargs:  Passed through to :meth:`.PropertyBase.__init__`.
                        If a ``default`` value is not provided, it is set
                        to something sensible.
        """

        default = kwargs.get('default', None)

        if default is None:
            if minval is not None and maxval is not None:
                default = (minval + maxval) / 2
            elif minval is not None:
                default = minval
            elif maxval is not None:
                default = maxval
            else:
                default = 0

        kwargs['default']    = default
        kwargs['minval']     = minval
        kwargs['maxval']     = maxval
        kwargs['clamped']    = clamped
        props.PropertyBase.__init__(self, **kwargs)


    def validate(self, instance, attributes, value):
        """Overrides :meth:`.PropertyBase.validate`. Validates the given
        number.

        Calls the :meth:`.PropertyBase.validate` method.
        Then, if the ``minval`` and/or ``maxval`` attributes have been set,
        and the given value is not within those values, a :exc:`ValueError` is
        raised.

        :param instance:   The owning :class:`.HasProperties` instance (or
                           ``None`` for unbound property values).

        :param attributes: Dictionary containing property attributes.

        :param value:      The value to validate.
        """

        props.PropertyBase.validate(self, instance, attributes, value)

        minval = attributes['minval']
        maxval = attributes['maxval']

        if minval is not None and value < minval:
            raise ValueError('Must be at least {}'.format(minval))

        if maxval is not None and value > maxval:
            raise ValueError('Must be at most {}'.format(maxval))


    def cast(self, instance, attributes, value):
        """Overrides :meth:`.PropertyBase.cast`.

        If the ``clamped`` attribute is ``True`` and the ``minval`` and/or
        ``maxval`` have been set, this function ensures that the given value
        lies within the ``minval`` and ``maxval`` limits. Otherwise the value
        is returned unchanged.
        """

        if value is None:
            return value

        clamped = attributes['clamped']

        if not clamped: return value

        minval = attributes['minval']
        maxval = attributes['maxval']

        if minval is not None and value < minval: return minval
        if maxval is not None and value > maxval: return maxval

        return value


class Int(Number):
    """A :class:`Number` which encapsulates an integer."""

    def __init__(self, **kwargs):
        """Create an ``Int`` property. """
        Number.__init__(self, **kwargs)


    def cast(self, instance, attributes, value):
        """Overrides :meth:`Number.cast`. Casts the given value to an ``int``,
        and then passes the value to :meth:`Number.cast`.
        """
        if value is None:
            return value
        return Number.cast(self, instance, attributes, int(value))


class Real(Number):
    """A :class:`.Number` which encapsulates a floating point number."""


    def __equals(self, a, b):
        """Custom equality function passed to :class`.PropertyBase.__init__`.

        Tests for equality according to the ``precision`` passed to
        :meth:`__init__`.
        """

        if any((a is None, b is None, self.__precision is None)):
            return a == b

        return abs(a - b) < self.__precision


    def __init__(self, precision=0.000000001, **kwargs):
        """Define a ``Real`` property.

        :param precision: Tolerance for equality testing. Set to ``None`` to
                          use exact equality.
        """
        self.__precision = precision

        Number.__init__(self, equalityFunc=self.__equals, **kwargs)


    def cast(self, instance, attributes, value):
        """Overrides :meth:`Number.cast`. Casts the given value to a ``float``,
        and then passes the value to :meth:`Number.cast`.
        """
        if value is None:
            return value
        return Number.cast(self, instance, attributes, float(value))


class Percentage(Real):
    """A :class:`Real` property which represents a percentage.

    A ``Percentage`` property is just a ``Real`` property with
    a default minimum value of ``0`` and default maximum value of ``100``.
    """

    def __init__(self, **kwargs):
        """Create a ``Percentage`` property."""
        kwargs['minval']  = kwargs.get('minval',    0.0)
        kwargs['maxval']  = kwargs.get('maxval',  100.0)
        kwargs['default'] = kwargs.get('default',  50.0)
        Real.__init__(self, **kwargs)


class String(props.PropertyBase):
    """A property which encapsulates a string."""

    def __init__(self, minlen=None, maxlen=None, **kwargs):
        """Cteate a ``String`` property.

        :param int minlen: Minimum valid string length.
        :param int maxlen: Maximum valid string length.
        """

        kwargs['default'] = kwargs.get('default', None)
        kwargs['minlen']  = minlen
        kwargs['maxlen']  = maxlen
        props.PropertyBase.__init__(self, **kwargs)


    def cast(self, instance, attributes, value):
        """Overrides :meth:`.PropertyBase.cast`.

        Casts the given value to a string. If the given value is the empty
        string, it is replaced with ``None``.
        """

        if value == '': return None
        else:           return value


    def validate(self, instance, attributes, value):
        """Overrides :meth:`.PropertyBase.validate`.

        Passes the given value to
        :meth:`.PropertyBase.validate`. Then, if either the
        ``minlen`` or ``maxlen`` attributes have been set, and the given
        value has length less than ``minlen`` or greater than ``maxlen``,
        raises a :exc:`ValueError`.
        """

        if value == '': value = None

        props.PropertyBase.validate(self, instance, attributes, value)

        if value is None: return

        if not isinstance(value, str):
            raise ValueError('Must be a string')

        minlen = attributes['minlen']
        maxlen = attributes['maxlen']

        if minlen is not None and len(value) < minlen:
            raise ValueError('Must have length at least {}'.format(minlen))

        if maxlen is not None and len(value) > maxlen:
            raise ValueError('Must have length at most {}'.format(maxlen))


class Choice(props.PropertyBase):
    """A property which may only be set to one of a set of predefined values.

    Choices can be added/removed via the :meth:`addChoice`,
    :meth:`removeChoice` method, and :meth:`setChoices` methods. Existing
    choices can be modified with the :meth:`updateChoice` method.

    Individual choices can be enabled/disabled via the :meth:`enableChoice`
    and :meth:`disableChoice` methods. The ``choiceEnabled`` attribute
    contains a dictionary of ``{choice : boolean}`` mappings
    representing the enabled/disabled state of each choice.

    A set of alternate values can be provided for each choice - these
    alternates will be accepted when assigning to a ``Choice`` property.

    .. note:: If you create a ``Choice`` property with non-string choice and
              alternate values, you may run into problems when using
              :mod:`.serialise` and/or :mod:`.cli` functionality, unless you
              set ``allowStr`` to ``True``.
    """

    def __init__(self,
                 choices=None,
                 alternates=None,
                 allowStr=False,
                 **kwargs):
        """Create a ``Choice`` property.

        :arg choices:    List of values, the possible values that this property
                         can take. Can alternately be a ``dict`` - see the note
                         above.

        :arg alternates: A list of lists, specificying alternate acceptable
                         values for each choice. Can also be a dict of
                         ``{choice : [alternates]}`` mappings. All alternate
                         values must be unique - different choices cannot have
                         equivalent alternate values.

        :arg allowStr:   If ``True``, string versions of any non-string choice
                         values will be accepted - ``str`` versions of each
                         choice are added as alternate values for that choice.
                         Defaults to ``False``.
        """

        if choices is None:
            choices    = []
            alternates = {}

        # Alternates are stored twice:
        #
        #   - As a dict of { choice    : [alternate] } mappings
        #   - As a dict of { alternate : choice      } mappings
        #
        # We generate the first dict here
        if alternates is None:
            alternates = {c : [] for c in choices}

        elif isinstance(alternates, dict):
            alternates = dict(alternates)

        elif isinstance(alternates, (list, tuple)):
            alternates = {c : list(a) for (c, a) in zip(choices, alternates)}

        # Add stringified versions of all
        # choices if allowStr is True
        if allowStr:
            for c in choices:
                strc = str(c)
                alts = alternates[c]
                if strc not in alts:
                    alts.append(strc)

        # Generate the second alternates dict
        altLists   = alternates
        alternates = self.__generateAlternatesDict(altLists)

        # Enabled flags are stored as a dict
        # of {choice : bool} mappings
        enabled = {choice: True for choice in choices}

        if len(choices) > 0: default = choices[0]
        else:                default = None

        if len(choices) != len(altLists):
            raise ValueError('Alternates are required for every choice')

        kwargs['choices']       = list(choices)
        kwargs['alternates']    = dict(alternates)
        kwargs['altLists']      = dict(altLists)
        kwargs['choiceEnabled'] = enabled
        kwargs['allowStr']      = allowStr
        kwargs['default']       = kwargs.get('default',      default)
        kwargs['allowInvalid']  = kwargs.get('allowInvalid', False)

        props.PropertyBase.__init__(self, **kwargs)


    def setDefault(self, default, instance=None):
        """Sets the default choice value. """
        if default not in self.getChoices(instance):
            raise ValueError('{} is not a choice'.format(default))

        self.setAttribute(instance, 'default', default)


    def enableChoice(self, choice, instance=None):
        """Enables the given choice. """
        choiceEnabled = dict(self.getAttribute(instance, 'choiceEnabled'))
        choiceEnabled[choice] = True
        self.setAttribute(instance, 'choiceEnabled', choiceEnabled)


    def disableChoice(self, choice, instance=None):
        """Disables the given choice. An attempt to set the property to
        a disabled value will result in a :exc:`ValueError`.
        """
        choiceEnabled = dict(self.getAttribute(instance, 'choiceEnabled'))
        choiceEnabled[choice] = False
        self.setAttribute(instance, 'choiceEnabled', choiceEnabled)


    def choiceEnabled(self, choice, instance=None):
        """Returns ``True`` if the given choice is enabled, ``False``
        otherwise.
        """
        return self.getAttribute(instance, 'choiceEnabled')[choice]


    def getChoices(self, instance=None):
        """Returns a list of the current choices. """
        return list(self.getAttribute(instance, 'choices'))


    def getAlternates(self, instance=None):
        """Returns a list of the current acceptable alternate values for each
        choice.
        """
        choices  = self.getAttribute(instance, 'choices')
        altLists = self.getAttribute(instance, 'altLists')

        return [altLists[c] for c in choices]


    def updateChoice(self,
                     choice,
                     newChoice=None,
                     newAlt=None,
                     instance=None):
        """Updates the choice value and/or alternates for the specified choice.
        """

        choices    = list(self.getAttribute(instance, 'choices'))
        altLists   = dict(self.getAttribute(instance, 'altLists'))
        idx        = choices.index(choice)

        if newChoice is not None:
            choices[ idx]       = newChoice
            altLists[newChoice] = altLists[choice]

            altLists.pop(choice)
        else:
            newChoice = choice

        if newAlt is not None: altLists[newChoice] = list(newAlt)

        self.__updateChoices(choices, altLists, instance)


    def setChoices(self, choices, alternates=None, instance=None):
        """Sets the list of possible choices (and their alternate values, if
        not None).
        """

        if alternates is None:
            alternates = {c : [] for c in choices}
        elif isinstance(alternates, (list, tuple)):
            alternates = {c : a for (c, a) in zip(choices, alternates)}
        elif isinstance(alternates, dict):
            alternates = dict(alternates)

        # Add stringified versions of all
        # choices if allowStr is True
        if self.getAttribute(instance, 'allowStr'):
            for c in choices:
                strc = str(c)
                alts = alternates[c]
                if strc not in alts:
                    alts.append(strc)

        if len(choices) != len(alternates):
            raise ValueError('Alternates are required for every choice')

        self.__updateChoices(choices, alternates, instance)


    def addChoice(self, choice, alternate=None, instance=None):
        """Adds a new choice to the list of possible choices."""

        if alternate is None: alternate = []
        else:                 alternate = list(alternate)

        choices  = list(self.getAttribute(instance, 'choices'))
        altLists = dict(self.getAttribute(instance, 'altLists'))

        if self.getAttribute(instance, 'allowStr'):
            strc = str(choice)
            if strc not in alternate:
                alternate.append(strc)

        choices.append(choice)
        altLists[choice] = list(alternate)

        self.__updateChoices(choices, altLists, instance)


    def removeChoice(self, choice, instance=None):
        """Removes the specified choice from the list of possible choices. """

        choices   = list(self.getAttribute(instance, 'choices'))
        altLists  = dict(self.getAttribute(instance, 'altLists'))

        choices .remove(choice)
        altLists.pop(   choice)

        self.__updateChoices(choices, altLists, instance)


    def __generateAlternatesDict(self, altLists):
        """Given a dictionary containing ``{choice : [alternates]}``
        mappings, creates and returns a dictionary containing
        ``{alternate : choice}`` mappings.

        Raises a ``ValueError`` if there are any duplicate alternate values.
        """

        alternates = {}

        for choice, altList in altLists.items():
            for alt in altList:
                if alt in alternates:
                    raise ValueError('Duplicate alternate value '
                                     '(choice: {}): {}'.format(choice, alt))
                alternates[alt] = choice

        return alternates


    def __updateChoices(self, choices, alternates, instance=None):
        """Used by all of the public choice modifying methods. Updates
        all choices, labels, and altenrates.

        :param choices:    A list of choice values

        :param alternates: A dict of ``{choice :  [alternates]}`` mappings.
        """

        propVal    = self.getPropVal(  instance)
        default    = self.getAttribute(instance, 'default')
        oldEnabled = self.getAttribute(instance, 'choiceEnabled')
        newEnabled = {}

        # Prevent notification while
        # we're updating constraints
        if propVal is not None:
            oldChoice  = propVal.get()
            notifState = propVal.getNotificationState()
            validState = propVal.allowInvalid()
            propVal.disableNotification()
            propVal.allowInvalid(True)

        for choice in choices:
            if choice in oldEnabled: newEnabled[choice] = oldEnabled[choice]
            else:                    newEnabled[choice] = True

        if default not in choices:
            default = choices[0]

        altLists   = alternates
        alternates = self.__generateAlternatesDict(altLists)

        self.setAttribute(instance, 'choiceEnabled', newEnabled)
        self.setAttribute(instance, 'altLists',      altLists)
        self.setAttribute(instance, 'alternates',    alternates)
        self.setAttribute(instance, 'choices',       choices)
        self.setAttribute(instance, 'default',       default)

        if propVal is not None:

            if oldChoice not in choices:
                if   default in choices: propVal.set(default)
                elif len(choices) > 0:   propVal.set(choices[0])
                else:                    propVal.set(None)

            propVal.setNotificationState(notifState)
            propVal.allowInvalid(        validState)

            if notifState:
                propVal.notifyAttributeListeners('choices', choices)

            if propVal.get() != oldChoice:
                propVal.propNotify()


    def validate(self, instance, attributes, value):
        """Overrides :meth:`.PropertyBase.validate`.

        Raises a :exc:`ValueError` if the given value is not one of the
        possible values for this :class:`Choice` property.
        """
        props.PropertyBase.validate(self, instance, attributes, value)

        choices    = self.getAttribute(instance, 'choices')
        enabled    = self.getAttribute(instance, 'choiceEnabled')
        alternates = self.getAttribute(instance, 'alternates')

        if len(choices) == 0: return

        # Check to see if this is an
        # acceptable alternate value
        altValue = alternates.get(value, None)

        if value not in choices and altValue not in choices:
            raise ValueError('Invalid choice ({})'    .format(value))

        if not enabled.get(value, False):
            raise ValueError('Choice is disabled ({})'.format(value))


    def cast(self, instance, attributes, value):
        """Overrides :meth:`.PropertyBase.cast`.

        Checks to see if the given value is a valid alternate value for a
        choice. If so, the alternate value is replaced with the choice value.
        """
        alternates = self.getAttribute(instance, 'alternates')
        return alternates.get(value, value)



class FilePath(String):
    """A property which represents a file or directory path.

    There is currently no support for validating a path which may be either a
    file or a directory - only one or the other.
    """

    def __init__(self, exists=False, isFile=True, suffixes=None, **kwargs):
        """Create a ``FilePath`` property.

        :param bool exists:   If ``True``, the path must exist.

        :param bool isFile:   If ``True``, the path must be a file. If
                              ``False``, the path must be a directory. This
                              check is only performed if ``exists`` is
                              ``True``.

        :param list suffixes: List of acceptable file suffixes (only relevant
                              if ``isFile`` is ``True``).
        """
        if suffixes is None:
            suffixes = []

        kwargs['exists']   = exists
        kwargs['isFile']   = isFile
        kwargs['suffixes'] = list(suffixes)

        String.__init__(self, **kwargs)


    def validate(self, instance, attributes, value):
        """Overrides :meth:`.PropertyBase.validate`.

        If the ``exists`` attribute is not ``True``, does nothing. Otherwise,
        if ``isFile`` is ``False`` and the given value is not a path to an
        existing directory, a :exc:`ValueError` is raised.

        If ``isFile`` is ``True``, and the given value is not a path to an
        existing file (which, if ``suffixes`` is not None, must end in one of
        the specified suffixes), a :exc:`ValueError` is raised.
        """

        String.validate(self, instance, attributes, value)

        exists   = attributes['exists']
        isFile   = attributes['isFile']
        suffixes = attributes['suffixes']

        if value is None: return
        if value == '':   return
        if not exists:    return

        if isFile:

            matchesSuffix = any([value.endswith(s) for s in suffixes])

            # If the file doesn't exist, it's bad
            if not op.isfile(value):
                raise ValueError('Must be a file ({})'.format(value))

            # if the file exists, and matches one of
            # the specified suffixes, then it's good
            if len(suffixes) == 0 or matchesSuffix: return

            # Otherwise it's bad
            else:
                raise ValueError(
                    'Must be a file ending in [{}] ({})'.format(
                        ','.join(suffixes), value))

        elif not op.isdir(value):
            raise ValueError('Must be a directory ({})'.format(value))


class List(props.ListPropertyBase):
    """A property which represents a list of items, of another property type.

    If you use ``List`` properties, you really should read the documentation
    for the :class:`.PropertyValueList`, as it contains important usage
    information.
    """

    def __init__(self, listType=None, minlen=None, maxlen=None, **kwargs):
        """Create a ``List`` property.

        :param listType:   A :class:`.PropertyBase` type, specifying the
                           values allowed in the list. If ``None``, anything
                           can be stored in the list, but no casting or
                           validation will occur.

        :param int minlen: Minimum list length.

        :param int maxlen: Maximum list length.
        """

        if (listType is not None) and \
           (not isinstance(listType, props.PropertyBase)):
            raise ValueError(
                'A list type (a PropertyBase instance) must be specified')

        kwargs['default'] = kwargs.get('default', [])
        kwargs['minlen']  = minlen
        kwargs['maxlen']  = maxlen

        # This needs to be removed when you update widgets_list.py
        self.embed = False

        props.ListPropertyBase.__init__(self, listType,  **kwargs)


    def validate(self, instance, attributes, value):
        """Overrides :meth:`.PropertyBase.validate`.

        Checks that the given value (which should be a list) meets the
        ``minlen``/``maxlen`` attribute. Raises a :exc:`ValueError` if it
        does not.
        """

        props.ListPropertyBase.validate(self, instance, attributes, value)

        minlen = attributes['minlen']
        maxlen = attributes['maxlen']

        if minlen is not None and len(value) < minlen:
            raise ValueError('Must have length at least {}'.format(minlen))
        if maxlen is not None and len(value) > maxlen:
            raise ValueError('Must have length at most {}'.format(maxlen))


class Colour(props.PropertyBase):
    """A property which represents a RGBA colour, stored as four floating
    point values in the range ``0.0 - 1.0``.

    Any value which can be interpreted by matplotlib as a RGB(A) colour is
    accepted. If an RGB colour is provided, the alpha channel is set to 1.0.
    """


    def __init__(self, **kwargs):
        """Create a ``Colour`` property.

        If the ``default`` ``kwarg`` is not provided, the default is set
        to white.
        """

        default = kwargs.get('default', (1.0, 1.0, 1.0, 1.0))

        if len(default) == 3:
            default = list(default) + [1.0]

        kwargs['default'] = default

        props.PropertyBase.__init__(self, **kwargs)


    def validate(self, instance, attributes, value):
        """Checks the given ``value``, and raises a :exc:`ValueError` if
        it does not consist of three or four floating point numbers in the
        range ``(0.0 - 1.0)``.
        """
        props.PropertyBase.validate(self, instance, attributes, value)
        mplcolors.to_rgba(value)


    def cast(self, instance, attributes, value):
        """Ensures that the given ``value`` contains three or four floating
        point numbers, in the range ``(0.0 - 1.0)``.

        If the alpha channel is not provided, it is set to the current alpha
        value (which defaults to ``1.0``).
        """
        if value is not None: return mplcolors.to_rgba(value)
        else:                 return value


class ColourMap(props.PropertyBase):
    """A property which encapsulates a :class:`matplotlib.colors.Colormap`.

    A ``ColourMap`` property can take any ``Colormap`` instance as its
    value. ColourMap values may be specified either as a
    ``Colormap`` instance, or as a string containing
    the name of a registered colour map instance.

    ``ColourMap`` properties also maintain an internal list of colour
    map names; while these names do not restrict the value that a ``ColourMap``
    property can take, they are used for display purposes - a widget which is
    created for a ``ColourMap`` instance will only display the options returned
    by the :meth:`getColourMaps` method. See the :func:`widgets._ColourMap`
    function.
    """

    def __init__(self, cmaps=None, **kwargs):
        """Define a ``ColourMap`` property. """

        default = kwargs.get('default', None)

        if cmaps is None:
            cmaps = []

        if default is None and len(cmaps) > 0:
            default = cmaps[0]

        kwargs['default'] = default
        kwargs['cmaps']   = list(cmaps)

        props.PropertyBase.__init__(self, **kwargs)


    def setColourMaps(self, cmaps, instance=None):
        """Set the colour maps for this property.

        :arg cmaps: a list of registered colour map names.
        """

        default = self.getAttribute(instance, 'default')

        if default not in cmaps:
            default = cmaps[0]

        self.setAttribute(instance, 'cmaps'  , cmaps)
        self.setAttribute(instance, 'default', default)


    def addColourMap(self, cmap, instance=None):
        """Add a colour map to the list.

        :arg cmap: The name of a registered colour map.
        """

        cmaps = self.getColourMaps(instance)

        if cmap not in cmaps:
            cmaps.append(cmap)
            self.setColourMaps(cmaps, instance)


    def getColourMaps(self, instance=None):
        """Returns a list containing the names of registered colour maps
        available for this property.
        """
        return list(self.getAttribute(instance, 'cmaps'))


    def validate(self, instance, attributes, value):
        """Overrides :meth:`.PropertyBase.validate`.

        Raises a :exc:`ValueError` if the given ``value`` is not a
        matplotlib :class:`.Colormap` instance.
        """

        if not isinstance(value, mplcolors.Colormap):
            raise ValueError('Colour map value is not a '
                             'matplotlib.colors.Colormap instance')


    def cast(self, instance, attributes, value):
        """Overrides :meth:`.PropertyBase.cast`.

        If the provided value is a string, an attempt is made to convert it to
        a colour map, via the ``matplotlib.colormaps`` registry. The value may
        either be the registered colour map name, or its ``Colormap.name``
        attribute. The match is case-insensitive.
        """

        if isinstance(value, str):

            # Case insensitive match against either
            # the registered colourmap key, or the
            # colourmap name. We allow any colour map
            # registered with matplotlib, but colour
            # maps added to this ColourMap property
            # are preferentially considered.
            cmapKeys   = list(self.getAttribute(instance, 'cmaps'))
            cmapKeys  += [c for c in mpl.colormaps.keys() if c not in cmapKeys]
            cmapNames  = [mpl.colormaps[cm].name for cm in cmapKeys]

            lCmapNames = [s.lower() for s in cmapNames]
            lCmapKeys  = [s.lower() for s in cmapKeys]
            value      = value.lower()
            idx        = None

            # Preferentially match against colour
            # map name rather than key. This is a
            # cheeky method of allowing us to
            # override built-in matplotlib colour
            # maps.
            try:               idx = lCmapNames.index(value)
            except ValueError: idx = lCmapKeys .index(value)

            if idx is None:
                choices = ','.join(cmapKeys)
                raise ValueError(f'Unknown colour map ({value}) - '
                                 f'valid choices are: {choices}')

            value = cmapKeys[idx]
            value = mpl.colormaps[value]

        return value


class BoundsValueList(propvals.PropertyValueList):
    """A :class:`.PropertyValueList` with values which represent bounds along
    a number of dimensions (up to 4).

    This class is used by the :class:`Bounds` property to encapsulate bounding
    values for an arbitrary number of dimensions. For ``N+1`` dimensions, the
    bounding values are stored as a list::

      [lo0, hi0, lo1, hi1, ..., loN, hiN]

    This class just adds some convenience methods and attributes to the
    ``PropertyValueList`` base class.  For a single dimension, a bound
    object has a ``lo`` value and a ``hi`` value, specifying the bounds along
    that dimension. To make things confusing, each dimension also has ``min``
    and ``max`` attributes, which define the minimum/maximum values that the
    ``lo`` and ``high`` values may take for that dimension.

    Some dynamic attributes are available on ``BoundsValueList`` objects,
    allowing access to and assignment of bound values and
    attributes. Dimensions ``0, 1, 2, 3`` respectively map to identifiers
    ``x, y, z, t``. If an attempt is made to access/assign an attribute
    corresponding to a dimension which does not exist on a particular
    ``BoundsValueList`` instance (e.g. attribute ``t`` on a 3-dimensional
    list), an :exc:`IndexError` will be raised. Here is an example of dynamic
    bound attribute access::

        class MyObj(props.HasProperties):
            myBounds = Bounds(ndims=4)

        obj = MyObj()

        # set/access lo/hi values together
        xlo, xhi = obj.mybounds.x
        obj.mybounds.z = (25, 30)

        # set/access lo/hi values separately
        obj.mybounds.xlo = 2
        obj.mybounds.zhi = 50

        # get the length of the bounds for a dimension
        ylen = obj.mybounds.ylen

        # set/access the minimum/maximum
        # constraints for a dimension
        obj.mybounds.xmin = 0
        tmax = obj.mybounds.tmax
    """


    def __init__(self, *args, **kwargs):
        """Create a ``BoundsValueList`` instance - see
        :meth:`.PropertyValueList.__init__`.
        """
        propvals.PropertyValueList.__init__(self, *args, **kwargs)


    def getLo(self, axis=None):
        """Return the low value for the given (0-indexed) axis. If ``axis`` is
        not specified, the low bounds for all axes are returned.
        """
        if axis is not None: return self[axis * 2]
        else:                return self[::2]


    def getHi(self, axis=None):
        """Return the high value for the given (0-indexed) axis. If ``axis``
        is not specified, the high bounds for all axes are returned.
        """
        if axis is not None: return self[axis * 2 + 1]
        else:                return self[1::2]


    def getRange(self, axis):
        """Return the (low, high) values for the given (0-indexed) axis."""
        return (self.getLo(axis), self.getHi(axis))


    def getLen(self, axis):
        """Return the distance between the low and high values for the
        specified axis.
        """
        return abs(self.getHi(axis) - self.getLo(axis))


    def setLimit(self, axis, limit, value):
        """Sets the value for the specified axis and limit
        (0 == low, 1 == high).
        """
        self[axis * 2 + limit] = value


    def getLimit(self, axis, limit):
        """Returns the value for the specified axis and limit
        (0 == low, 1 == high).
        """
        return self[axis * 2 + limit]


    def setLo(self, axis, value):
        """Set the low value for the specified axis."""
        self[axis * 2] = value


    def setHi(self, axis, value):
        """Set the high value for the specified axis."""
        self[axis * 2 + 1] = value


    def setRange(self, axis, loval, hival):
        """Set the low and high values for the specified axis."""
        self[axis * 2:axis * 2 + 2] = [loval, hival]


    def getMin(self, axis):
        """Return the minimum value (the low limit) for the specified axis."""
        return self.getPropertyValueList()[axis * 2].getAttribute('minval')


    def getMax(self, axis):
        """Return the maximum value (the high limit) for the specified axis."""
        return self.getPropertyValueList()[axis * 2 + 1].getAttribute('maxval')


    def setMin(self, axis, value):
        """Set the minimum value for the specified axis."""
        self.getPropertyValueList()[axis * 2]    .setAttribute('minval', value)
        self.getPropertyValueList()[axis * 2 + 1].setAttribute('minval', value)


    def setMax(self, axis, value):
        """Set the maximum value for the specified axis."""
        self.getPropertyValueList()[axis * 2]    .setAttribute('maxval', value)
        self.getPropertyValueList()[axis * 2 + 1].setAttribute('maxval', value)


    def getLimits(self, axis):
        """Return (minimum, maximum) limit values for the specified axis."""
        return (self.getMin(axis), self.getMax(axis))


    def setLimits(self, axis, minval, maxval):
        """Set the minimum and maximum limit values for the specified axis."""
        self.setMin(axis, minval)
        self.setMax(axis, maxval)


    def inBounds(self, point):
        """Returns ``True`` if the given point (a sequence of numbers) lies
        within the bounds represented by this ``BoundsValueList``, ``False``
        otherwise.
        """
        if 2 * len(point) != len(self):
            raise ValueError('Invalid number of dimensions: {}'.format(point))

        for ax, coord in enumerate(point):
            if coord < self.getLo(ax) or coord > self.getHi(ax):
                return False

        return True


    def __getattr__(self, name):
        """Return the specified value. Raises an :exc:`AttributeError` for
        unrecognised attributes, or an :exc:`IndexError` if an attempt is made
        to access bound values values of a higher dimension than this list
        contains.
        """

        lname = name.lower()

        # TODO this is easy to read, but
        # could be made much more efficient
        if   lname == 'x':    return self.getRange( 0)
        elif lname == 'y':    return self.getRange( 1)
        elif lname == 'z':    return self.getRange( 2)
        elif lname == 't':    return self.getRange( 3)
        elif lname == 'lo':   return self.getLo()
        elif lname == 'hi':   return self.getHi()
        elif lname == 'xlo':  return self.getLo(    0)
        elif lname == 'xhi':  return self.getHi(    0)
        elif lname == 'ylo':  return self.getLo(    1)
        elif lname == 'yhi':  return self.getHi(    1)
        elif lname == 'zlo':  return self.getLo(    2)
        elif lname == 'zhi':  return self.getHi(    2)
        elif lname == 'tlo':  return self.getLo(    3)
        elif lname == 'thi':  return self.getHi(    3)
        elif lname == 'xlen': return self.getLen(   0)
        elif lname == 'ylen': return self.getLen(   1)
        elif lname == 'zlen': return self.getLen(   2)
        elif lname == 'tlen': return self.getLen(   3)
        elif lname == 'xmin': return self.getMin(   0)
        elif lname == 'ymin': return self.getMin(   1)
        elif lname == 'zmin': return self.getMin(   2)
        elif lname == 'tmin': return self.getMin(   3)
        elif lname == 'xmax': return self.getMax(   0)
        elif lname == 'ymax': return self.getMax(   1)
        elif lname == 'zmax': return self.getMax(   2)
        elif lname == 'tmax': return self.getMax(   3)
        elif lname == 'xlim': return self.getLimits(0)
        elif lname == 'ylim': return self.getLimits(1)
        elif lname == 'zlim': return self.getLimits(2)
        elif lname == 'tlim': return self.getLimits(3)

        raise AttributeError('{} has no attribute called {}'.format(
            self.__class__.__name__, name))


    def __setattr__(self, name, value):
        """Set the specified value. Raises an :exc:`IndexError` if an attempt
        is made to assign bound values values of a higher dimension than this
        list contains.
        """

        lname = name.lower()

        if   lname == 'x':    self.setRange( 0, *value)
        elif lname == 'y':    self.setRange( 1, *value)
        elif lname == 'z':    self.setRange( 2, *value)
        elif lname == 't':    self.setRange( 3, *value)
        elif lname == 'xlo':  self.setLo(    0, value)
        elif lname == 'xhi':  self.setHi(    0, value)
        elif lname == 'ylo':  self.setLo(    1, value)
        elif lname == 'yhi':  self.setHi(    1, value)
        elif lname == 'zlo':  self.setLo(    2, value)
        elif lname == 'zhi':  self.setHi(    2, value)
        elif lname == 'tlo':  self.setLo(    3, value)
        elif lname == 'thi':  self.setHi(    3, value)
        elif lname == 'xmin': self.setMin(   0, value)
        elif lname == 'ymin': self.setMin(   1, value)
        elif lname == 'zmin': self.setMin(   2, value)
        elif lname == 'tmin': self.setMin(   3, value)
        elif lname == 'xmax': self.setMax(   0, value)
        elif lname == 'ymax': self.setMax(   1, value)
        elif lname == 'zmax': self.setMax(   2, value)
        elif lname == 'tmax': self.setMax(   3, value)
        elif lname == 'xlim': self.setLimits(0, *value)
        elif lname == 'ylim': self.setLimits(1, *value)
        elif lname == 'zlim': self.setLimits(2, *value)
        elif lname == 'tlim': self.setLimits(3, *value)
        else:                 self.__dict__[name] = value


class Bounds(List):
    """A property which represents numeric bounds in any number of dimensions,
    as long as that number is no more than 4.

    ``Bounds`` values are stored in a :class:`BoundsValueList`, a list of
    integer or floating point values, with two values (lo, hi) for each
    dimension.

    ``Bounds`` values may also have bounds of their own, i.e. minimium/maximum
    values that the bound values can take. These bound-limits are referred to
    as 'min' and 'max', and can be set via the :meth:`BoundsValueList.setMin`
    and :meth:`BoundsValueList.setMax` methods. The advantage to using these
    methods, instead of using, for example,
    :meth:`.PropertyValue.setAttribute`, is that if you use the latter you will
    have to set the attributes on both the low and the high values.
    """

    def __init__(self,
                 ndims=1,
                 real=True,
                 minDistance=None,
                 clamped=True,
                 minval=None,
                 maxval=None,
                 **kwargs):
        """Create a ``Bounds`` property.

        :arg ndims:        Number of dimensions. This is (currently)
                           not a property attribute, hence it cannot
                           be changed.

        :arg real:         If ``True`` (the default), the bound values are
                           stored as :class:`Real` values; otherwise, they
                           are stored as :class:`Int` values.

        :arg minDistance:  Minimum distance to be maintained between the
                           low/high values for each dimension.

        :arg clamped:      If ``True`` (the default), the bound values are
                           clamped to their limits. See the :class:`Number`
                           class.

        :arg minval:       Initial minimum value to use for all dimensions

        :arg maxval:       Initial maximum value to use for all dimensions
        """

        default = kwargs.get('default', None)

        if minDistance is None:
            minDistance = 0.0

        if default is None:
            default = [0.0, minDistance] * ndims
            if minval is not None:
                default[0::2] = [minval] * ndims
            if maxval is not None:
                default[1::2] = [maxval] * ndims

        if ndims < 1 or ndims > 4:
            raise ValueError('Only bounds of one to four '
                             'dimensions are supported')

        if len(default) != 2 * ndims:
            raise ValueError('{} bound values are required'.format(2 * ndims))

        kwargs['default']     = default
        kwargs['minDistance'] = minDistance
        kwargs['minval']      = minval
        kwargs['maxval']      = maxval

        self._real   = real
        self._ndims  = ndims

        if real: listType = Real(clamped=clamped)
        else:    listType = Int( clamped=clamped)

        List.__init__(self,
                      listType=listType,
                      minlen=ndims * 2,
                      maxlen=ndims * 2,
                      **kwargs)


    def _makePropVal(self, instance):
        """Overrides :meth:`.ListPropertyBase._makePropVal`.

        Creates and returns a ``BoundsValueList`` instead of a
        ``PropertyValueList``, so callers get to use the convenience
        methods/attributes defined in the BVL class.
        """

        default = self.getAttribute(None, 'default', None)
        minval  = self.getAttribute(None, 'minval',  None)
        maxval  = self.getAttribute(None, 'maxval',  None)
        bvl     = BoundsValueList(
            instance,
            name=self.getLabel(instance),
            values=default,
            itemCastFunc=self._listType.cast,
            itemEqualityFunc=self._listType._equalityFunc,
            itemValidateFunc=self._listType.validate,
            listValidateFunc=self.validate,
            listAttributes=self._defaultAttributes,
            itemAttributes=self._listType._defaultAttributes)

        for i in range(self._ndims):
            if minval is not None:
                bvl.setMin(i, minval)
            if maxval is not None:
                bvl.setMax(i, maxval)

        return bvl


    def validate(self, instance, attributes, value):
        """Overrides :meth:`.PropertyBase.validate`.

        Raises a :exc:`ValueError` if the given value (a list of min/max
        values) is of the wrong length or data type, or if any of the min
        values are greater than the corresponding max value.
        """

        minDistance = attributes['minDistance']

        # the List.validate method will check
        # the value length and type for us
        List.validate(self, instance, attributes, value)

        for i in range(self._ndims):

            imin = value[i * 2]
            imax = value[i * 2 + 1]

            if imin > imax:
                raise ValueError('Minimum bound must be smaller '
                                 'than maximum bound (dimension {}, '
                                 '{} - {}'.format(i, imin, imax))

            if imax - imin < minDistance:
                raise ValueError('Minimum and maximum bounds must be at '
                                 'least {} apart'.format(minDistance))


class PointValueList(propvals.PropertyValueList):
    """A list of values which represent a point in some n-dimensional (up to 4)
    space.

    This class is used by the :class:`Point` property to encapsulate point
    values for between 1 and 4 dimensions.

    This class just adds some convenience methods and attributes to the
    :class:`.PropertyValueList` base class, in a similar manner to the
    :class:`BoundsValueList` class.

    The point values for each dimension may be queried/assigned via the
    dynamic attributes ``x, y, z, t``, which respectively map to dimensions
    ``0, 1, 2, 3``. When querying/assigning point values, you may use
    `GLSL-like swizzling
    <http://www.opengl.org/wiki/Data_Type_(GLSL)#Swizzling>`_. For example::

        class MyObj(props.HasProperties):
            mypoint = props.Point(ndims=3)

        obj = MyObj()

        y,z = obj.mypoint.yz

        obj.mypoint.zxy = (3,6,1)
    """


    def __init__(self, *args, **kwargs):
        """Create a ``PointValueList`` - see
        :meth:`.PropertyValueList.__init__`.
        """
        propvals.PropertyValueList.__init__(self, *args, **kwargs)


    def getPos(self, axis):
        """Return the point value for the specified (0-indexed) axis."""
        return self[axis]


    def setPos(self, axis, value):
        """Set the point value for the specified axis."""
        self[axis] = value


    def getMin(self, axis):
        """Get the minimum limit for the specified axis."""
        return self.getPropertyValueList()[axis].getAttribute('minval')


    def getMax(self, axis):
        """Get the maximum limit for the specified axis."""
        return self.getPropertyValueList()[axis].getAttribute('maxval')


    def getLimits(self, axis):
        """Get the (minimum, maximum) limits for the specified axis."""
        return (self.getMin(axis), self.getMax(axis))


    def setMin(self, axis, value):
        """Set the minimum limit for the specified axis."""
        self.getPropertyValueList()[axis].setAttribute('minval', value)


    def setMax(self, axis, value):
        """Set the maximum limit for the specified axis."""
        self.getPropertyValueList()[axis].setAttribute('maxval', value)


    def setLimits(self, axis, minval, maxval):
        """Set the minimum and maximum limits for the specified axis."""
        self.setMin(axis, minval)
        self.setMax(axis, maxval)


    def __getattr__(self, name):
        """Return the specified point value. Raises an :exc:`AttributeError`
        for unrecognised attributes, or an :exc:`IndexError` if a dimension
        which does not exist for this ``PointValueList`` is specified.
        """

        lname = name.lower()

        if any([dim not in 'xyzt' for dim in lname]):
            raise AttributeError('{} has no attribute called {}'.format(
                self.__class__.__name__, name))

        res = []
        for dim in lname:
            if   dim == 'x': res.append(self[0])
            elif dim == 'y': res.append(self[1])
            elif dim == 'z': res.append(self[2])
            elif dim == 't': res.append(self[3])

        if len(res) == 1: return res[0]
        return res


    def __setattr__(self, name, value):
        """Set the specified point value. Raises an :exc:`IndexError` if a
        dimension which does not exist for this ``PointValueList`` is
        specified.
        """

        lname = name.lower()

        if any([dim not in 'xyzt' for dim in lname]):
            self.__dict__[name] = value
            return

        if len(lname) == 1:
            value = [value]

        if len(lname) != len(value):
            raise AttributeError('Improper number of values '
                                 '({}) for attribute {}'.format(
                                     len(value), lname))

        newvals = self[:]

        for dim, val in zip(lname, value):
            if   dim == 'x': newvals[0] = val
            elif dim == 'y': newvals[1] = val
            elif dim == 'z': newvals[2] = val
            elif dim == 't': newvals[3] = val

        self[:] = newvals


class Point(List):
    """A property which represents a point in some n-dimensional (up to 4)
    space.

    ``Point`` property values are stored in a :class:`PointValueList`, a
    list of integer or floating point values, one for each dimension.
    """

    def __init__(self,
                 ndims=2,
                 real=True,
                 **kwargs):
        """Create a ``Point`` property.

        :param int ndims: Number of dimensions.

        :param bool real: If ``True`` the point values are stored as
                          :class:`Real` values, otherwise they are
                          stored as :class:`Int` values.
        """

        default = kwargs.get('default', None)

        if default is None: default = [0] * ndims

        if real:
            default = [float(v) for v in default]

        if ndims < 1 or ndims > 4:
            raise ValueError('Only points of one to four '
                             'dimensions are supported')

        elif len(default) != ndims:
            raise ValueError('{} point values are required'.format(ndims))

        kwargs['default']    = default

        self._ndims   = ndims
        self._real    = real

        if real: listType = Real(clamped=True)
        else:    listType = Int( clamped=True)

        List.__init__(self,
                      listType=listType,
                      minlen=ndims,
                      maxlen=ndims,
                      **kwargs)


    def _makePropVal(self, instance):
        """Overrides :meth:`.ListPropertyBase._makePropVal`.

        Creates and returns a ``PointValueList`` instead of a
        ``PropertyValueList``, so callers get to use the convenience
        methods/attributes defined in the PVL class.
        """

        default = self.getAttribute(None, 'default', None)

        pvl = PointValueList(
            instance,
            name=self.getLabel(instance),
            values=default,
            itemCastFunc=self._listType.cast,
            itemEqualityFunc=self._listType._equalityFunc,
            itemValidateFunc=self._listType.validate,
            listValidateFunc=self.validate,
            listAttributes=self._defaultAttributes,
            itemAttributes=self._listType._defaultAttributes)

        return pvl


class ArrayProxy(propvals.PropertyValue):
    """A proxy class indended to encapsulate a ``numpy`` array.  ``ArrayProxy``
    instances are used by the :class:`Array` property type.


    An ``ArrayProxy`` is a :class:`.PropertyValue` which contains, and tries
    to act like, a ``numpy`` array.  Element access andassignment, and
    attribute access can be performed through the ``ArrayProxy``.


    All element assignments which occur via an ``ArrayProxy`` instance will
    result in notification of all registered listeners (see the
    :meth:`.PropertyValue.addListener` method). A limitation of this
    implementation is that notification will occur even if the assignment
    does not change the array values.


    The underlying ``numpy`` array is accessible via the :meth:`getArray`
    method. However, changes made directly to the numpy array will
    bypass the :class:`.PropertyValue` notification procedure.
    """


    def __init__(self, *args, **kwargs):
        """Create an ``ArrayProxy``. All arguments are passed through to the
        :meth:`.PropertyValue.__init__` method.
        """

        def defaultEquals(this, other):
            if isinstance(this,  ArrayProxy): this  = this .getArray()
            if isinstance(other, ArrayProxy): other = other.getArray()

            return np.all(this == other)

        kwargs['equalityFunc'] = kwargs.get('equalityFunc', defaultEquals)
        propvals.PropertyValue.__init__(self, *args, **kwargs)


    def get(self):
        """Overrides :meth:`.PropertyValue.get`. Returns this ``ArrayProxy``.
        """
        return self


    def getArray(self):
        """Returns a reference to the ``numpy`` array encapsulated by this
        ``ArrayProxy``.
        """
        return propvals.PropertyValue.get(self)


    def __getattr__(self, name):
        """Returns the attribute of the ``numpy`` array with the given name.
        """
        return getattr(self.getArray(), name)


    def __getitem__(self, *args, **kwargs):
        """Calls the ``__getitem``__ method of the ``numpy`` array. """
        return self.getArray().__getitem__(*args, **kwargs)


    def __setitem__(self, *args, **kwargs):
        """Calls the ``__setitem__`` method of the ``numpy`` array, and triggers
        notification of all registered listeners.
        """

        array = self.getArray()

        array.__setitem__(*args, **kwargs)

        notifState = self.getNotificationState()
        self.disableNotification()

        self.set(array)

        self.setNotificationState(notifState)
        self.propNotify()


class Array(props.PropertyBase):
    """A property which represents a ``numpy`` array. Each array is
    encapsulated within an :class:`ArrayProxy` instance.
    """

    def __init__(self, dtype=None, shape=None, resizable=True, **kwargs):
        """Create an ``Array`` property.

        .. warning:: If you set a ``default`` value here (see
                     :meth:`.PropertyBase.__init__`), *do not* use a ``numpy``
                     array - use a regular python list. It will be converted
                     in a ``numpy`` array internally. An equality test is made
                     on the ``default`` attribute, so if you use a ``numpy``
                     array, a :exc:`ValueError` will be raised, as ``numpy``
                     performs equality tests on an element-wise basis.

        :arg dtype:     ``numpy`` data type.

        :arg shape:     Initial shape of the array.

        :arg resizable: Defaults to ``True``. If ``False``, the array size
                        will be fixed to the ``shape`` specified
                        here. Different sized arrays will still be allowed, if
                        the ``allowInvalid`` parameter to
                        :meth:`.PropertyBase.__init__` is set to ``True``.
        """

        if dtype is None: dtype = np.float64
        if shape is None: shape = (4, 4)

        kwargs['dtype']     = dtype
        kwargs['shape']     = shape
        kwargs['resizable'] = resizable
        kwargs['default']   = kwargs.get('default', np.zeros(shape, dtype))

        props.PropertyBase.__init__(self, **kwargs)


    def _makePropVal(self, instance):
        """Overrides :meth`.PropertyBase._makePropVal`. Creates and returns
        an :class:`.ArrayProxy` for the given ``instance``.
        """

        default = self.getAttribute(None, 'default', None)
        ap      = ArrayProxy(
            instance,
            name=self.getLabel(instance),
            value=default,
            castFunc=self.cast,
            validateFunc=self.validate,
            allowInvalid=self._allowInvalid,
            **self._defaultAttributes)

        return ap


    def cast(self, instance, attributes, value):
        """Overrides :meth`.PropertyBase.cast`. Casts the given value to a
        ``numpy`` array (with the data type that was specified in
        :meth:`__init__`).
        """
        dtype = attributes['dtype']
        return np.array(value, dtype=dtype)


    def validate(self, instance, attributes, value):
        """Overrides :meth`.PropertyBase.validate`. If the given ``value`` has
        the wrong data type, or this ``Array`` is not resizable and the
        ``value`` has the wrong shape, a :exc:`ValueError` is raised.
        """
        props.PropertyBase.validate(self, instance, attributes, value)

        dtype     = attributes['dtype']
        shape     = attributes['shape']
        resizable = attributes['resizable']

        if value.dtype != dtype:
            raise ValueError('Invalid data type: {} (should be {})'.format(
                value.dtype, dtype))

        if (not resizable) and (value.shape != shape):
            raise ValueError('Invalid shape: {} (should be {})'.format(
                value.shape, shape))



#!/usr/bin/env python
#
# __init__.py - Sets up the fsleyes_props package namespace.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
# flake8: noqa
"""``fsleyes_props`` is a framework for event-driven programming using python
descriptors, similar in functionality to, and influenced by `Enthought Traits
<http://code.enthought.com/projects/traits/>`_.

-------------
Example usage
-------------

::

    >>> import fsleyes_props as props

    >>> class PropObj(props.Props):
            myProperty = props.Boolean()

    >>> myPropObj = PropObj()


    # Access the property value as a normal attribute:
    >>> myPropObj.myProperty = True
    >>> myPropObj.myProperty
    True


    # access the props.Boolean instance:
    >>> myPropObj.getProp('myProperty')
    <props.prop.Boolean at 0x1045e2710>


    # access the underlying props.PropertyValue object
    >>> myPropObj.getPropVal('myProperty')
    <props.prop.PropertyValue instance at 0x1047ef518>


    # Receive notification of property value changes
    >>> def myPropertyChanged(value, *args):
            print('New property value: {}'.format(value))

    >>> myPropObj.addListener(
           'myProperty', 'myListener', myPropertyChanged)

    >>> myPropObj.myProperty = False
    New property value: False


    # Remove a previously added listener
    >>> myPropObj.removeListener('myListener')


-----------------
Package structure
-----------------


To use ``fsleyes_props``, your first step will be to define a subclass of
:class:`.HasProperties` (a.k.a. :class:`Props`), which contains one or more
:class:`.PropertyBase` class attributes (see the :mod:`.properties_types`
module for the available types).


Once you have an instance of your ``HasProperties`` class, you can then create
a GUI for it using the functions defined in the :mod:`.build` and
:mod:`.widgets` modules, and the GUI specification building blocks defined in
the :mod:`build_parts` module. You can also generate a command-line interface
using the functions defined in the :mod:`.cli` module.


All of the classes and functions referred to above are available in the
``fsleyes_props`` namespace, so you only need to ``import fsleyes_props`` to
access them. You will however need to call the :func:`initGUI` function if you
want to use any of the GUI generation functionality, before they are made
available at the ``fsleyes_props`` namespace level.


---------------
Boring overview
---------------


Lots of the code in this package is probably very confusing. First of all, you
will need to understand python descriptors.  Descriptors are a way of adding
properties to python objects, and allowing them to be accessed as if they were
just simple attributes of the object, but controlling the way that the
attributes are accessed and assigned.


The following link provides a good overview, and contains the ideas
which form the basis for the implementation in this package:


 -  http://nbviewer.ipython.org/urls/gist.github.com/\
ChrisBeaumont/5758381/raw/descriptor_writeup.ipynb


And if you've got 30 minutes, this video gives a very good
introduction to descriptors:


 - http://pyvideo.org/video/1760/encapsulation-with-descriptors


A :class:`.HasProperties` subclass contains a collection of
:class:`.PropertyBase` instances as class attributes. When an instance of the
``HasProperties`` class is created, a :class:`.PropertyValue` object is
created for each of the ``PropertyBase`` instances (or a
:class:`~.PropertyValueList` for :class:`.ListPropertyBase` instances).  Each
of these ``PropertyValue`` instances encapsulates a single value, of any type
(a ``PropertyValueList`` instance encapsulates multiple ``PropertyValue``
instances).  Whenever this value changes, the ``PropertyValue`` instance
notifies any registered listeners of the change.


^^^^^^^^^^^^
Notification
^^^^^^^^^^^^


Application code may be notified of property changes by registering a callback
listener on a ``PropertyValue`` object, via the equivalent methods:

  - :meth:`.HasProperties.addListener`
  - :meth:`.PropertyBase.addListener`
  - :meth:`.PropertyValue.addListener`

Such a listener will be notified of changes to the ``PropertyValue`` object
managed by the ``PropertyBase`` object, and associated with the
``HasProperties`` instance. For ``ListPropertyBase`` properties, a listener
registered through one of the above methods will be notified of changes to the
entire list.  Alternately, a listener may be registered with individual items
contained in the list (see :meth:`.PropertyValueList.getPropertyValueList`).


^^^^^^^^^^
Validation
^^^^^^^^^^


When a ``PropertyValue`` accepts a new value, it passes the value to the
:meth:`.PropertyBase.validate` method of its parent ``PropertyBase`` instance
to determine whether the new value is valid.  The ``PropertyValue`` object
may allow its underlying value to be set to something invalid, but it will
tell registered listeners whether the new value is valid or
invalid. ``PropertyValue`` objects can alternately be configured to raise a
:exc:`ValueError` on an attempt to set them to an invalid value, but this has
some caveats - see the ``PropertyValue`` documentation. Finally, to make things
more confusing, some ``PropertyBase`` types will configure their
``PropertyValue`` objects to perform implicit casts when the property value is
set.


The default validation logic of most ``PropertyBase`` objects can be
configured via *attributes*. For example, the :class:`.Number` property
allows ``minval`` and ``maxval`` attributes to be set.  These may be set via
``PropertyBase`` constructors, (i.e. when it is defined as a class attribute
of a ``HasProperties`` definition), and may be queried and changed on
individual ``HasProperties`` instances via the
:meth:`.HasProperties.getAttribute`/:meth:`.HasProperties.setAttribute`
methods; similarly named methods are also available on ``PropertyBase``
instances. Some ``PropertyBase`` classes provide additional convenience
methods for accessing their attributes (e.g. :meth`.Choice.addChoice`).


^^^^^^^^^^^^^^^^^^^^^^^^^^^
Binding and Synchronisation
^^^^^^^^^^^^^^^^^^^^^^^^^^^


Properties from different ``HasProperties`` instances may be bound to each
other, so that changes in one are propagated to the other - see the
:mod:`.bindable` module.  Building on this is the :mod:`.syncable` module and
its :class:`.SyncableHasProperties` class, which allows a one-to-many (one
parent, multiple children) synchronisation hierarchy to be maintained, whereby
all the properties of a child instance are by default synchronised to those of
the parent, and this synchronisation can be independently enabled/disabled for
each property. To use this functionality, simply inherit from the
``SyncableHasProperties`` class instead of the ``HasProperties`` class.


------------
API overview
------------


^^^^^^^^^^^^^^
Property types
^^^^^^^^^^^^^^


The following classes are provided as building-blocks for your application
code:

.. autosummary::
   :nosignatures:

   ~fsleyes_props.properties.HasProperties
   ~fsleyes_props.syncable.SyncableHasProperties
   ~fsleyes_props.properties_types.Object
   ~fsleyes_props.properties_types.Boolean
   ~fsleyes_props.properties_types.Int
   ~fsleyes_props.properties_types.Real
   ~fsleyes_props.properties_types.Percentage
   ~fsleyes_props.properties_types.String
   ~fsleyes_props.properties_types.FilePath
   ~fsleyes_props.properties_types.Choice
   ~fsleyes_props.properties_types.List
   ~fsleyes_props.properties_types.Colour
   ~fsleyes_props.properties_types.ColourMap
   ~fsleyes_props.properties_types.Bounds
   ~fsleyes_props.properties_types.Point


^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
Command line and string serialisation
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


The following functions are provided to manage command-line argument
generation and parsing:

.. autosummary::
   :nosignatures:

   ~fsleyes_props.cli.applyArguments
   ~fsleyes_props.cli.addParserArguments
   ~fsleyes_props.cli.generateArguments


The following functions are provided for serialisation/deserialisation of
property values to/from strings (equivalent methods are also available on
:class:`.HasProperties` instances):

.. autosummary::
   :nosignatures:

   ~fsleyes_props.serialise.serialise
   ~fsleyes_props.serialise.deserialise


^^^^^^^^^^^^^^^^^^^^^^^^^^^^
GUI specification/generation
^^^^^^^^^^^^^^^^^^^^^^^^^^^^


The following classes are provided for you to create GUI specifications:

.. autosummary::
   :nosignatures:

   ~fsleyes_props.build_parts.ViewItem
   ~fsleyes_props.build_parts.Button
   ~fsleyes_props.build_parts.Toggle
   ~fsleyes_props.build_parts.Label
   ~fsleyes_props.build_parts.Widget
   ~fsleyes_props.build_parts.Group
   ~fsleyes_props.build_parts.NotebookGroup
   ~fsleyes_props.build_parts.HGroup
   ~fsleyes_props.build_parts.VGroup


If the :func:`initGUI` function is called, the following GUI-related functions
will be made available in the ``fsleyes_props`` package namespace:


.. autosummary::
   :nosignatures:

   ~fsleyes_props.widgets.makeWidget
   ~fsleyes_props.widgets.makeListWidget
   ~fsleyes_props.widgets.makeListWidgets
   ~fsleyes_props.widgets.makeSyncWidget
   ~fsleyes_props.widgets.bindWidget
   ~fsleyes_props.widgets.unbindWidget
   ~fsleyes_props.widgets.bindListWidgets
   ~fsleyes_props.build.buildGUI
   ~fsleyes_props.build.buildDialog


^^^^^^^^^^^^^
Miscellaneous
^^^^^^^^^^^^^

The :func:`.suppress` module provides some context managers allowing
notification of properties to be suppressed in a ``with`` statement.
"""


__version__ = '1.9.5'


import sys
import logging


log = logging.getLogger(__name__)


from .properties import (
    PropertyOwner,
    HasProperties,
    HasProps,
    Props,
    DisabledError)

from .properties_value import (
    safeCall
)

from .bindable import (
    bindPropVals,
    propValsAreBound,
    Bidict)

from .properties_types import (
    Object,
    Boolean,
    Int,
    Real,
    Percentage,
    String,
    FilePath,
    Choice,
    List,
    Colour,
    ColourMap,
    Bounds,
    Point,
    Array)

from .syncable import (
    SyncError,
    SyncableHasProperties,)

from .cli import (
    SkipArgument,
    applyArguments,
    addParserArguments,
    generateArguments)

from .serialise import (
    serialise,
    deserialise)

from .build_parts import (
    ViewItem,
    Button,
    Toggle,
    Label,
    Widget,
    Group,
    NotebookGroup,
    HGroup,
    VGroup)

from .suppress import (
    suppress,
    suppressAll,
    skip)


from .cache import (
    PropCache,
    CacheError)


def initGUI():
    """If you wish to use GUI generation functionality, calling this function
    will add the relevant functions to the ``fsleyes_props`` package namespace.
    """

    mod = sys.modules[__name__]

    from . import widgets

    # These properties belong in the
    # widgets module, but are complex
    # enough to get their own modules.
    # We monkey patch the widgets
    # module here (rather than widgets
    # itself) to avoid circular
    # dependency problems
    from .widgets_list    import _List
    from .widgets_bounds  import _Bounds
    from .widgets_point   import _Point
    from .widgets_choice  import _Choice
    from .widgets_boolean import _Boolean
    from .widgets_number  import _Number

    widgets._List    = _List
    widgets._Bounds  = _Bounds
    widgets._Point   = _Point
    widgets._Choice  = _Choice
    widgets._Boolean = _Boolean
    widgets._Number  = _Number

    from .widgets import (
        makeWidget,
        makeListWidget,
        makeListWidgets,
        makeSyncWidget,
        bindWidget,
        unbindWidget,
        bindListWidgets)

    from .build import (
        buildGUI,
        buildDialog)

    mod.makeWidget      = makeWidget
    mod.makeListWidget  = makeListWidget
    mod.makeListWidgets = makeListWidgets
    mod.makeSyncWidget  = makeSyncWidget
    mod.bindWidget      = bindWidget
    mod.unbindWidget    = unbindWidget
    mod.bindListWidgets = bindListWidgets
    mod.buildGUI        = buildGUI
    mod.buildDialog     = buildDialog



#!/usr/bin/env python
#
# test_syncable.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import pytest

import fsleyes_props as props


def test_basic_usage():
    class Foo(props.SyncableHasProperties):
        myint = props.Int(default=0)

    p = Foo()
    c = Foo(parent=p)

    assert c.isSyncedToParent('myint')
    assert c.anySyncedToParent()
    assert c.allSyncedToParent()
    assert c.canBeSyncedToParent('myint')
    assert c.canBeUnsyncedFromParent('myint')

    assert list(p.getChildren()) == [c]
    assert c.getParent()         ==  p

    assert p.myint == 0
    assert c.myint == 0

    p.myint = 10

    assert p.myint == 10
    assert c.myint == 10

    c.myint = 20

    assert p.myint == 20
    assert c.myint == 20


def test_unsync():
    class Foo(props.SyncableHasProperties):
        myint = props.Int(default=0)

    changeCalled = [False]

    def syncChange(*a):
        changeCalled[0] = True

    p = Foo()
    c = Foo(parent=p)

    p.myint = 10
    assert p.myint == 10
    assert c.myint == 10

    c.addSyncChangeListener('myint', 'syncChange', syncChange)

    c.unsyncFromParent('myint')

    assert changeCalled[0]
    changeCalled[0] = False

    p.myint = 20
    assert p.myint == 20
    assert c.myint == 10

    c.syncToParent('myint')

    assert changeCalled[0]
    assert c.myint == 20

    changeCalled[0] = False
    c.removeSyncChangeListener('myint', 'syncChange')

    p.myint = 30
    assert not changeCalled[0]
    assert p.myint == 30
    assert c.myint == 30


def test_unsync_all():
    class Foo(props.SyncableHasProperties):
        int1 = props.Int(default=0)
        int2 = props.Int(default=0)

    p = Foo()
    c = Foo(parent=p)

    p.int1, p.int2 = 10, 20

    assert (c.int1, c.int2) == (10, 20)

    c.unsyncAllFromParent()

    p.int1, p.int2 = 30, 40

    assert c.int1, c.int2 == (10, 20)

    p.int1, p.int2 = 50, 60

    c.syncAllToParent()
    assert (c.int1, c.int2) == (50, 60)



def test_bindingDirection():

    class Foo(props.SyncableHasProperties):
        myint = props.Int(default=0)

    p = Foo()
    c = Foo(parent=p, direction=False)
    assert not c.getBindingDirection('myint')

    p.myint = 10

    c.unsyncFromParent('myint')
    p.myint = 20
    c.myint = 40

    assert p.myint == 20
    assert c.myint == 40

    c.syncToParent('myint')

    assert p.myint == 40
    assert c.myint == 40

    c.unsyncFromParent('myint')
    c.setBindingDirection(True, 'myint')

    p.myint = 75

    c.syncToParent('myint')
    assert p.myint == 75
    assert c.myint == 75


def test_detach():
    class Foo(props.SyncableHasProperties):
        myint = props.Int(default=0)

    p = Foo()
    c = Foo(parent=p)

    p.myint = 25
    assert c.myint == 25

    c.detachFromParent('myint')

    p.myint = 75
    assert c.myint == 25

    c = Foo(parent=p)

    p.myint = 15
    assert c.myint == 15

    c.detachAllFromParent()

    p.myint = 23
    assert c.myint == 15



def test_syncabe_list_link1():  _test_syncabe_list_link(1)
def test_syncabe_list_link2():  _test_syncabe_list_link(2)
def test_syncabe_list_link3():  _test_syncabe_list_link(3)
def test_syncabe_list_link4():  _test_syncabe_list_link(4)
def test_syncabe_list_link5():  _test_syncabe_list_link(5)
def test_syncabe_list_link6():  _test_syncabe_list_link(6)
def test_syncabe_list_link7():  _test_syncabe_list_link(7)
def test_syncabe_list_link8():  _test_syncabe_list_link(8)
def test_syncabe_list_link9():  _test_syncabe_list_link(9)
def test_syncabe_list_link10(): _test_syncabe_list_link(10)

# List properties whcih are synced together, and also
# where, within each instance, individual list items
# are bound across properties
def _test_syncabe_list_link(nchildren):
    class Thing(props.SyncableHasProperties):
        crange = props.Bounds(ndims=1, clamped=False)
        drange = props.Bounds(ndims=1, clamped=False)
        linklo = props.Boolean(default=True)

        def __init__(self, *args, **kwargs):
            props.SyncableHasProperties.__init__(self, *args, **kwargs)
            parent = self.getParent()

            if parent is not None:
                self.addListener('linklo',
                                 str(id(self)),
                                 self.__linkloChanged)
                if self.linklo:  self.__linkloChanged()

        def __linkloChanged(self, *a):
            self.__updateLink(self.linklo, 0)

        def __updateLink(self, val, idx):

            drangePV = self.drange.getPropertyValueList()[idx]
            crangePV = self.crange.getPropertyValueList()[idx]

            if props.propValsAreBound(drangePV, crangePV) == val:
                return

            props.bindPropVals(drangePV,
                               crangePV,
                               bindval=True,
                               bindatt=False,
                               unbind=not val)

            if val:
                crangePV.set(drangePV.get())

    parent   = Thing()
    children = [Thing(parent=parent) for i in range(nchildren)]

    parent.crange = [10, 20]
    parent.drange = [5,  10]

    for i in [parent] + children:
        assert i.crange == [5, 20]
        assert i.drange == [5, 10]


def test_syncable_class_hierarchy():

    class Boo(props.SyncableHasProperties):
        mybool = props.Boolean()

    class Foo(Boo):
        myint = props.Int()

    bp = Boo()
    bc = Boo(parent=bp)

    bp.mybool = False
    assert(not bc.mybool)
    bp.mybool = True
    assert(bc.mybool)
    bc.mybool = False
    assert(not bp.mybool)
    bc.mybool = True
    assert(bp.mybool)

    fp = Foo()
    fc = Foo(parent=fp)

    fp.mybool = False
    assert(not fc.mybool)
    fp.mybool = True
    assert(fc.mybool)
    fc.mybool = False
    assert(not fp.mybool)
    fc.mybool = True
    assert(fp.mybool)


def test_nobind_nounbind():
    class Foo(props.SyncableHasProperties):
        int1 = props.Int()
        int2 = props.Int()
        int3 = props.Int()

    p = Foo(nobind=['int1'], nounbind=['int2'])
    c = Foo(nobind=['int1'], nounbind=['int2'], parent=p)

    assert not c.canBeSyncedToParent('int1')
    assert     c.canBeSyncedToParent('int2')
    assert     c.canBeSyncedToParent('int3')
    assert     c.canBeUnsyncedFromParent('int1')
    assert not c.canBeUnsyncedFromParent('int2')
    assert     c.canBeUnsyncedFromParent('int3')
    assert not c.isSyncedToParent('int1')
    assert     c.isSyncedToParent('int2')
    assert     c.isSyncedToParent('int3')
    assert     c.anySyncedToParent()
    assert not c.allSyncedToParent()

    p.int1 = 123
    c.int1 = 456
    assert (p.int1, c.int1) == (123, 456)

    p.int2 = 123
    c.int2 = 456
    assert (p.int2, c.int2) == (456, 456)

    p.int3 = 123
    c.int3 = 456
    assert (p.int3, c.int3) == (456, 456)

    with pytest.raises(props.SyncError):
        c.syncToParent('int1')

    with pytest.raises(props.SyncError):
        c.unsyncFromParent('int2')

    c.detachAllFromParent()

    c.int1, c.int2, c.int3 = 10, 20, 30
    p.int1, p.int2, p.int3 = 40, 50, 60

    assert (c.int1, c.int2, c.int3) == (10, 20, 30)
    assert (p.int1, p.int2, p.int3) == (40, 50, 60)



#!/usr/bin/env python
#
# test_widget_point.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import wx
import numpy as np

import fsleyes_props               as props
import fsleyes_widgets.floatslider as floatslider


from . import (run_with_wx, MockMouseEvent, addall, realYield)


def setup_module():
    props.initGUI()


class MyObj(props.HasProperties):

    mypointi = props.Point(ndims=2, real=False)
    mypointf = props.Point(ndims=2)


def  test_widget_point(): run_with_wx(_test_widget_point)
def _test_widget_point(parent):

    sim = wx.UIActionSimulator()
    obj = MyObj()

    obj.mypointi.setLimits(0, 0, 100)
    obj.mypointi.setLimits(1, 0, 100)
    obj.mypointf.setLimits(0, 0, 100)
    obj.mypointf.setLimits(1, 0, 100)

    obj.mypointi.x = 25
    obj.mypointi.y = 75
    obj.mypointf.x = 20
    obj.mypointf.y = 80

    mypointi = props.makeWidget(parent, obj, 'mypointi', showLimits=False)
    mypointf = props.makeWidget(parent, obj, 'mypointf', showLimits=False)

    xi, yi = mypointi.GetChildren()
    xf, yf = mypointf.GetChildren()

    for s in [xi, yi, xf, yf]:
        assert isinstance(s, floatslider.SliderSpinPanel)

    assert xi.GetValue() == 25
    assert yi.GetValue() == 75
    assert xf.GetValue() == 20
    assert yf.GetValue() == 80

    addall(parent, [mypointi, mypointf])

    tests = [
        (xi, 'mypointi', 'x'),
        (yi, 'mypointi', 'y'),
        (xf, 'mypointf', 'x'),
        (yf, 'mypointf', 'y')]


    for widget, prop, att in tests:
        val = np.random.randint(0, 100)
        setattr(getattr(obj, prop), att, val)
        realYield(2)
        assert np.isclose(widget.GetValue(), val)

        val = np.random.randint(0, 100)

        ev = MockMouseEvent(widget.slider, (val / 100.0, 0.5))
        widget.slider._FloatSlider__onMouseDown(ev)
        widget.slider._FloatSlider__onMouseUp(ev)
        realYield()

        assert abs(getattr(getattr(obj, prop), att) - val) < 10
